from typing import Callable, List, Optional

import jax
import jax.numpy as jnp

from e3nn_jax import Irreps, IrrepsData
from e3nn_jax.util.decorators import overload_for_irreps_without_data


def normalize_function(phi):
    with jax.ensure_compile_time_eval():
        k = jax.random.PRNGKey(0)
        x = jax.random.normal(k, (1_000_000,))
        c = jnp.mean(phi(x) ** 2) ** 0.5

        if jnp.allclose(c, 1.0):
            return phi
        else:

            def rho(x):
                return phi(x) / c

            return rho


def parity_function(phi):
    with jax.ensure_compile_time_eval():
        x = jnp.linspace(0.0, 10.0, 256)

        a1, a2 = phi(x), phi(-x)
        if jnp.max(jnp.abs(a1 - a2)) < 1e-5:
            return 1
        elif jnp.max(jnp.abs(a1 + a2)) < 1e-5:
            return -1
        else:
            return 0


def is_zero_in_zero(phi):
    with jax.ensure_compile_time_eval():
        return jnp.allclose(phi(jnp.array(0.0)), 0.0)


@overload_for_irreps_without_data(irrepsdata_argnums=[0])
def scalar_activation(input: IrrepsData, acts: List[Optional[Callable[[float], float]]]) -> IrrepsData:
    assert isinstance(input, IrrepsData)

    assert len(input.irreps) == len(acts), (input.irreps, acts)

    list = []

    irreps_out = []
    for (mul, (l_in, p_in)), x, act in zip(input.irreps, input.list, acts):
        if act is not None:
            if l_in != 0:
                raise ValueError(
                    f"Activation: cannot apply an activation function to a non-scalar input. {input.irreps} {acts}"
                )

            act = normalize_function(act)

            p_out = parity_function(act) if p_in == -1 else p_in
            if p_out == 0:
                raise ValueError(
                    "Activation: the parity is violated! The input scalar is odd but the activation is neither even nor odd."
                )

            irreps_out.append((mul, (0, p_out)))
            if x is None:
                if is_zero_in_zero(act):
                    list.append(None)
                else:
                    list.append(act(jnp.ones(input.shape + (mul, 1))))
            else:
                list.append(act(x))
        else:
            irreps_out.append((mul, (l_in, p_in)))
            list.append(x)

    irreps_out = Irreps(irreps_out)

    if acts and acts.count(acts[0]) == len(acts):
        # for performance, if all the activation functions are the same, we can apply it to the contiguous array as well
        contiguous = input.contiguous if acts[0] is None else normalize_function(acts[0])(input.contiguous)
        return IrrepsData(irreps=irreps_out, contiguous=contiguous, list=list)

    return IrrepsData.from_list(irreps_out, list, input.shape)


# TODO remove this class and follow the same pattern as scalar_activation
class KeyValueActivation:
    irreps_key: Irreps
    irreps_value: Irreps
    irreps_out: Irreps

    def __init__(self, irreps_key, irreps_value, phi):
        self.irreps_key = Irreps(irreps_key)
        self.irreps_value = Irreps(irreps_value)

        # TODO compute irreps_out
        # irreps_out =

    def __call__(self, keys, values):
        return [jax.vmap(key_value_activation)(self.phi, k, v) for k, v in zip(keys, values)]


def key_value_activation(phi, key, value):
    assert key.ndim == 1
    assert value.ndim == 1

    d = value.shape[0]
    key = key / jnp.sqrt(1 / 16 + jnp.sum(key**2))  # 1/16 is arbitrary small... but not too small...
    scalar = jnp.sum(key * value)
    scalar = normalize_function(phi)(scalar)
    return d**0.5 * scalar * key  # component normalized



from functools import partial
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
from e3nn_jax import IrrepsData, index_add
from e3nn_jax.util import prod
from jax import lax


@partial(jax.jit, static_argnums=(1, 2, 3, 4))
def lowpass_filter(input, scale, strides, transposed=False, steps=(1, 1, 1)):
    r"""Lowpass filter for 3D field.

    Args:
        input: [..., x, y, z]
        scale (float): typically 2.0
        strides (int): typically 1 or 2
        transposed (bool): if True, dilate the input instead of stride
        steps (tuple): physical dimensions of the voxel grid

    Known issues:
           stride=2    transposed=True
        64 -------> 32 --------------> 63
    """

    if isinstance(strides, int):
        strides = (strides,) * 3

    with jax.ensure_compile_time_eval():
        sigma = 0.5 * (scale**2 - 1) ** 0.5

        size = int(1 + 2 * 2.5 * sigma)
        if size % 2 == 0:
            size += 1

        r = jnp.linspace(-1, 1, size)
        x = r * steps[0] / min(steps)
        x = x[jnp.abs(x) <= 1]
        y = r * steps[1] / min(steps)
        y = y[jnp.abs(y) <= 1]
        z = r * steps[2] / min(steps)
        z = z[jnp.abs(z) <= 1]
        lattice = jnp.stack(jnp.meshgrid(x, y, z, indexing="ij"), axis=-1)  # [x, y, z, R^3]
        lattice = (size // 2) * lattice

        kernel = jnp.exp(-jnp.sum(lattice**2, axis=-1) / (2 * sigma**2))
        kernel = kernel / jnp.sum(kernel)

        if transposed:
            kernel = kernel * strides[0] * strides[1] * strides[2]

        kernel = kernel[None, None]  # [1, 1, x, y, z]

    if scale <= 1:
        assert strides == (1,) * 3
        return input

    pad = (kernel.shape[-3] // 2, kernel.shape[-2] // 2, kernel.shape[-1] // 2)

    output = input
    output = output.reshape(-1, 1, *output.shape[-3:])
    output = lax.conv_general_dilated(
        lhs=output,
        rhs=kernel,
        lhs_dilation=strides if transposed else (1, 1, 1),
        window_strides=(1, 1, 1) if transposed else strides,
        padding=((pad[0], pad[0]), (pad[1], pad[1]), (pad[2], pad[2])),
        dimension_numbers=("NCXYZ", "IOXYZ", "NCXYZ"),
    )

    output = output.reshape(*input.shape[:-3], *output.shape[-3:])
    return output


def interpolate_bilinear(input: jnp.ndarray, x: float, y: float, z: float) -> jnp.ndarray:
    r"""interpolate voxels in coordinate (x, y, z).

    Args:
        input: [..., x, y, z]
        x: x coordinate
        y: y coordinate
        z: z coordinate
    """
    # based on http://stackoverflow.com/a/12729229
    x_lo = jnp.floor(x).astype(int)
    x_hi = x_lo + 1
    y_lo = jnp.floor(y).astype(int)
    y_hi = y_lo + 1
    z_lo = jnp.floor(z).astype(int)
    z_hi = z_lo + 1

    nx, ny, nz = input.shape[-3:]

    def xclip(x):
        return jnp.clip(x, 0, nx - 1)

    def yclip(y):
        return jnp.clip(y, 0, ny - 1)

    def zclip(z):
        return jnp.clip(z, 0, nz - 1)

    Ia = input[..., xclip(x_lo), yclip(y_lo), zclip(z_lo)]
    Ib = input[..., xclip(x_hi), yclip(y_lo), zclip(z_lo)]
    Ic = input[..., xclip(x_lo), yclip(y_hi), zclip(z_lo)]
    Id = input[..., xclip(x_hi), yclip(y_hi), zclip(z_lo)]
    Ie = input[..., xclip(x_lo), yclip(y_lo), zclip(z_hi)]
    If = input[..., xclip(x_hi), yclip(y_lo), zclip(z_hi)]
    Ig = input[..., xclip(x_lo), yclip(y_hi), zclip(z_hi)]
    Ih = input[..., xclip(x_hi), yclip(y_hi), zclip(z_hi)]

    wa = (x_hi - x) * (y_hi - y) * (z_hi - z)
    wb = (x - x_lo) * (y_hi - y) * (z_hi - z)
    wc = (x_hi - x) * (y - y_lo) * (z_hi - z)
    wd = (x - x_lo) * (y - y_lo) * (z_hi - z)
    we = (x_hi - x) * (y_hi - y) * (z - z_lo)
    wf = (x - x_lo) * (y_hi - y) * (z - z_lo)
    wg = (x_hi - x) * (y - y_lo) * (z - z_lo)
    wh = (x - x_lo) * (y - y_lo) * (z - z_lo)

    return wa * Ia + wb * Ib + wc * Ic + wd * Id + we * Ie + wf * If + wg * Ig + wh * Ih


@partial(jax.jit, static_argnums=(1,))
def _zoom(
    input: jnp.ndarray,
    output_size: Tuple[int, int, int],
) -> jnp.ndarray:
    nx, ny, nz = input.shape[-3:]

    def f(n_src, n_dst):
        # new pixel positions in source coordinate
        a = n_src / n_dst * jnp.arange(n_dst)
        delta = 0.5 * (n_src / n_dst - 1)
        return delta + a

    xi = f(nx, output_size[0])
    yi = f(ny, output_size[1])
    zi = f(nz, output_size[2])

    xg, yg, zg = jnp.meshgrid(xi, yi, zi, indexing="ij")

    output = jax.vmap(interpolate_bilinear, (None, 0, 0, 0), -1)(input, xg.flatten(), yg.flatten(), zg.flatten())
    output = output.reshape(*input.shape[:-3], *output_size)
    return output


def zoom(
    input: jnp.ndarray,
    *,
    resize_rate: Optional[Tuple[float, float, float]] = None,
    output_size: Optional[Tuple[int, int, int]] = None,
) -> jnp.ndarray:
    r"""Rescale a 3D image by bilinear interpolation.

    Args:
        input: array of shape ``[..., x, y, z]``
        resize_rate: tuple of 3 floats
        output_size: tuple of 3 ints

    Returns:
        3D image of size output_size
    """
    nx, ny, nz = input.shape[-3:]

    if resize_rate is not None:
        assert output_size is None

        if isinstance(resize_rate, (float, int)):
            resize_rate = (resize_rate,) * 3

        output_size = (round(nx * resize_rate[0]), round(ny * resize_rate[1]), round(nz * resize_rate[2]))

    assert isinstance(output_size, tuple)

    return _zoom(input, output_size)


def _index_max_norm(input, strides):
    norms = jnp.sum(input**2, axis=-1)
    shape = input.shape[:-1]
    assert len(shape) == len(strides)
    idxs = jnp.arange(prod(shape)).reshape(shape)

    def g(a, b):
        an, ai = a
        bn, bi = b
        which = an >= bn
        return (jnp.where(which, an, bn), jnp.where(which, ai, bi))

    _, idxs = lax.reduce_window(
        (norms, idxs),
        (-jnp.inf, -1),
        g,
        window_dimensions=strides,
        window_strides=strides,
        padding=((0, 0),) * len(strides),
    )
    return idxs


@jax.custom_vjp
def norm_maxpool(input, strides):
    idxs = _index_max_norm(input, strides)
    return input.reshape(prod(input.shape[:-1]), input.shape[-1])[idxs]


def norm_maxpool_fwd(input, strides):
    idxs = _index_max_norm(input, strides)
    output = input.reshape(prod(input.shape[:-1]), input.shape[-1])[idxs]
    return output, (idxs, input.shape)


def norm_maxpool_bwd(residuals, grad):
    idxs, shape = residuals
    idxs = idxs.flatten()
    grad = index_add(idxs, grad.reshape(idxs.shape[0], -1), prod(shape[:-1]))
    grad = grad.reshape(shape)
    return (grad, None)


norm_maxpool.defvjp(norm_maxpool_fwd, norm_maxpool_bwd)


def maxpool(input: IrrepsData, strides) -> IrrepsData:
    r"""
    Args:
        input: IrrepsData of shape [x, y, z]
        strides: tuple of ints

    Returns:
        IrrepsData
    """
    assert isinstance(input, IrrepsData)
    assert len(input.shape) == len(strides)

    list = [None if x is None else jax.vmap(lambda x: norm_maxpool(x, strides), -2, -2)(x) for x in input.list]

    shape = tuple(a // s for a, s in zip(input.shape, strides))

    return IrrepsData.from_list(input.irreps, list, shape)



import math

import numpy as np


def su2_generators(j: float) -> np.ndarray:
    """
    Return the generators of the SU(2) group of dimension `2j+1`.
    """
    m = np.arange(-j, j)
    raising = np.diag(-np.sqrt(j * (j + 1) - m * (m + 1)), k=-1)

    m = np.arange(-j + 1, j + 1)
    lowering = np.diag(np.sqrt(j * (j + 1) - m * (m - 1)), k=1)

    m = np.arange(-j, j + 1)
    return np.stack(
        [
            0.5 * (raising + lowering),  # x (usually)
            np.diag(1j * m),  # z (usually)
            -0.5j * (raising - lowering),  # -y (usually)
        ],
        axis=0,
    )


def su2_clebsch_gordan(j1: float, j2: float, j3: float) -> np.ndarray:
    """Calculates the Clebsch-Gordon matrix
    for SU(2) coupling j1 and j2 to give j3.
    Parameters
    ----------
    j1 : float
        Total angular momentum 1.
    j2 : float
        Total angular momentum 2.
    j3 : float
        Total angular momentum 3.
    Returns
    -------
    cg_matrix : numpy.array
        Requested Clebsch-Gordan matrix.
    """
    assert isinstance(j1, (int, float))
    assert isinstance(j2, (int, float))
    assert isinstance(j3, (int, float))
    mat = np.zeros((int(2 * j1 + 1), int(2 * j2 + 1), int(2 * j3 + 1)))
    if int(2 * j3) in range(int(2 * abs(j1 - j2)), int(2 * (j1 + j2)) + 1, 2):
        for m1 in (x / 2 for x in range(-int(2 * j1), int(2 * j1) + 1, 2)):
            for m2 in (x / 2 for x in range(-int(2 * j2), int(2 * j2) + 1, 2)):
                if abs(m1 + m2) <= j3:
                    mat[int(j1 + m1), int(j2 + m2), int(j3 + m1 + m2)] = _su2_cg((j1, m1), (j2, m2), (j3, m1 + m2))
    return mat / math.sqrt(2 * j3 + 1)


def _su2_cg(idx1, idx2, idx3):
    """Calculates the Clebsch-Gordon coefficient
    for SU(2) coupling (j1,m1) and (j2,m2) to give (j3,m3).
    Parameters
    ----------
    j1 : float
        Total angular momentum 1.
    j2 : float
        Total angular momentum 2.
    j3 : float
        Total angular momentum 3.
    m1 : float
        z-component of angular momentum 1.
    m2 : float
        z-component of angular momentum 2.
    m3 : float
        z-component of angular momentum 3.
    Returns
    -------
    cg_coeff : float
        Requested Clebsch-Gordan coefficient.
    """
    from fractions import Fraction

    j1, m1 = idx1
    j2, m2 = idx2
    j3, m3 = idx3

    if m3 != m1 + m2:
        return 0
    vmin = int(max([-j1 + j2 + m3, -j1 + m1, 0]))
    vmax = int(min([j2 + j3 + m1, j3 - j1 + j2, j3 + m3]))

    def f(n):
        assert n == round(n)
        return math.factorial(round(n))

    C = (
        (2.0 * j3 + 1.0)
        * Fraction(
            f(j3 + j1 - j2) * f(j3 - j1 + j2) * f(j1 + j2 - j3) * f(j3 + m3) * f(j3 - m3),
            f(j1 + j2 + j3 + 1) * f(j1 - m1) * f(j1 + m1) * f(j2 - m2) * f(j2 + m2),
        )
    ) ** 0.5

    S = 0
    for v in range(vmin, vmax + 1):
        S += (-1.0) ** (v + j2 + m2) * Fraction(
            f(j2 + j3 + m1 - v) * f(j1 - m1 + v), f(v) * f(j3 - j1 + j2 - v) * f(j3 + m3 - v) * f(v + j1 - j2 - m3)
        )
    C = C * S
    return C



from typing import Any, List, NamedTuple, Optional, Tuple, Union

import haiku as hk
import jax
import jax.numpy as jnp

from e3nn_jax import Irreps, IrrepsData

from ._core_tensor_product import _sum_tensors


class Instruction(NamedTuple):
    i_in: int
    i_out: int
    path_shape: tuple
    path_weight: float


class FunctionalLinear:
    irreps_in: Irreps
    irreps_out: Irreps
    instructions: List[Instruction]
    output_mask: jnp.ndarray

    def __init__(
        self,
        irreps_in: Any,
        irreps_out: Any,
        instructions: Optional[List[Tuple[int, int]]] = None,
        biases: Optional[Union[List[bool], bool]] = None,
        path_normalization: str = "element",
    ):
        assert path_normalization in ["element", "path"]

        irreps_in = Irreps(irreps_in)
        irreps_out = Irreps(irreps_out)

        if instructions is None:
            # By default, make all possible connections
            instructions = [
                (i_in, i_out)
                for i_in, (_, ir_in) in enumerate(irreps_in)
                for i_out, (_, ir_out) in enumerate(irreps_out)
                if ir_in == ir_out
            ]

        instructions = [
            Instruction(
                i_in=i_in,
                i_out=i_out,
                path_shape=(irreps_in[i_in].mul, irreps_out[i_out].mul),
                path_weight=1,
            )
            for i_in, i_out in instructions
        ]

        def alpha(ins):
            x = sum(
                irreps_in[i.i_in if path_normalization == "element" else ins.i_in].mul
                for i in instructions
                if i.i_out == ins.i_out
            )
            return 1.0 if x == 0 else x

        instructions = [
            Instruction(i_in=ins.i_in, i_out=ins.i_out, path_shape=ins.path_shape, path_weight=alpha(ins) ** (-0.5))
            for ins in instructions
        ]

        if biases is None:
            biases = len(irreps_out) * (False,)
        if isinstance(biases, bool):
            biases = [biases and ir.is_scalar() for _, ir in irreps_out]

        assert len(biases) == len(irreps_out)
        assert all(ir.is_scalar() or (not b) for b, (_, ir) in zip(biases, irreps_out))

        instructions += [
            Instruction(i_in=-1, i_out=i_out, path_shape=(mul_ir.dim,), path_weight=1.0)
            for i_out, (bias, mul_ir) in enumerate(zip(biases, irreps_out))
            if bias
        ]

        with jax.ensure_compile_time_eval():
            if irreps_out.dim > 0:
                output_mask = jnp.concatenate(
                    [
                        jnp.ones(mul_ir.dim)
                        if any((ins.i_out == i_out) and (0 not in ins.path_shape) for ins in instructions)
                        else jnp.zeros(mul_ir.dim)
                        for i_out, mul_ir in enumerate(irreps_out)
                    ]
                )
            else:
                output_mask = jnp.ones(0)

        self.irreps_in = irreps_in
        self.irreps_out = irreps_out
        self.instructions = instructions
        self.output_mask = output_mask

    def aggregate_paths(self, paths, output_shape) -> IrrepsData:
        output = [
            _sum_tensors(
                [out for ins, out in zip(self.instructions, paths) if ins.i_out == i_out],
                shape=output_shape
                + (
                    mul_ir_out.mul,
                    mul_ir_out.ir.dim,
                ),
                empty_return_none=True,
            )
            for i_out, mul_ir_out in enumerate(self.irreps_out)
        ]
        return IrrepsData.from_list(self.irreps_out, output, output_shape)

    def __call__(self, ws: List[jnp.ndarray], input: IrrepsData) -> IrrepsData:
        input = IrrepsData.new(self.irreps_in, input)
        assert len(input.shape) == 0

        paths = [
            ins.path_weight * w
            if ins.i_in == -1
            else (None if input.list[ins.i_in] is None else ins.path_weight * jnp.einsum("uw,ui->wi", w, input.list[ins.i_in]))
            for ins, w in zip(self.instructions, ws)
        ]
        return self.aggregate_paths(paths, input.shape)

    def matrix(self, ws: List[jnp.ndarray]) -> jnp.ndarray:
        r"""
        Compute the matrix representation of the linear operator.

        Args:
            ws: List of weights.

        Returns:
            The matrix representation of the linear operator. The matrix is shape ``(irreps_in.dim, irreps_out.dim)``.
        """
        output = jnp.zeros((self.irreps_in.dim, self.irreps_out.dim))
        for ins, w in zip(self.instructions, ws):
            assert ins.i_in != -1
            mul_in, ir_in = self.irreps_in[ins.i_in]
            mul_out, ir_out = self.irreps_out[ins.i_out]
            output = output.at[self.irreps_in.slices()[ins.i_in], self.irreps_out.slices()[ins.i_out]].add(
                ins.path_weight
                * jnp.einsum("uw,ij->uiwj", w, jnp.eye(ir_in.dim)).reshape((mul_in * ir_in.dim, mul_out * ir_out.dim))
            )
        return output


class Linear(hk.Module):
    def __init__(self, irreps_out, *, irreps_in=None):
        super().__init__()

        self.irreps_out = Irreps(irreps_out)
        self.irreps_in = Irreps(irreps_in) if irreps_in is not None else None
        self.instructions = None
        self.biases = False

    def __call__(self, input):
        if self.irreps_in is None and not isinstance(input, IrrepsData):
            raise ValueError("the input of Linear must be an IrrepsData, or `irreps_in` must be specified")
        if self.irreps_in is not None:
            input = IrrepsData.new(self.irreps_in, input)

        input = input.remove_nones().simplify()
        lin = FunctionalLinear(input.irreps, self.irreps_out, self.instructions, biases=self.biases)
        w = [
            hk.get_parameter(
                f"b[{ins.i_out}] {lin.irreps_out[ins.i_out]}", shape=ins.path_shape, init=hk.initializers.Constant(0.0)
            )
            if ins.i_in == -1
            else hk.get_parameter(
                f"w[{ins.i_in},{ins.i_out}] {lin.irreps_in[ins.i_in]},{lin.irreps_out[ins.i_out]}",
                shape=ins.path_shape,
                init=hk.initializers.RandomNormal(),
            )
            for ins in lin.instructions
        ]
        return lin(w, input)



from functools import partial

import jax
import jax.numpy as jnp

from e3nn_jax import IrrepsData, elementwise_tensor_product, scalar_activation
from e3nn_jax.util.decorators import overload_for_irreps_without_data


@partial(jax.jit, static_argnums=(1, 2, 3, 4))
def _gate(input: IrrepsData, even_act, odd_act, even_gate_act, odd_gate_act) -> IrrepsData:
    scalars, gated = input, None
    for j, (_, ir) in enumerate(input.irreps):
        if ir.l > 0:
            scalars, gated = input.split([j])
            break
    assert scalars.irreps.lmax == 0

    # No gates:
    if gated is None:
        return scalar_activation(scalars, [even_act if ir.p == 1 else odd_act for _, ir in scalars.irreps])

    # Get the scalar gates:
    gates = None
    for i in range(j + 1):
        if scalars.irreps[i:].num_irreps == gated.irreps.num_irreps:
            scalars, gates = scalars.split([i])
            break

    if gates is None:
        raise ValueError(
            f"Gate: did not manage to split the input {input.irreps} into scalars, gates ({scalars.irreps}) and gated "
            f"({gated.irreps})."
        )

    scalars = scalar_activation(scalars, [even_act if ir.p == 1 else odd_act for _, ir in scalars.irreps])
    gates = scalar_activation(gates, [even_gate_act if ir.p == 1 else odd_gate_act for _, ir in gates.irreps])

    return IrrepsData.cat([scalars, elementwise_tensor_product(gates, gated)])


@overload_for_irreps_without_data((0,))
def gate(input: IrrepsData, even_act=None, odd_act=None, even_gate_act=None, odd_gate_act=None) -> IrrepsData:
    r"""Gate activation function.

    The input is split into scalars that are activated separately, scalars that are used as gates, and non-scalars that are
    multiplied by the gates.

    List of assumptions:

    - The scalars are on the left side of the input.
    - The gate scalars are on the right side of the scalars.

    Args:
        input (IrrepsData): Input data.
        acts: The list of activation functions. Its length must be equal to the number of scalar blocks in the input.

    Returns:
        IrrepsData: Output data.

    Examples:
        The 3 even scalars are used as gates.
        >>> gate("12x0e + 3x0e + 2x1e + 1x2e")
        12x0e+2x1e+1x2e

        Odd scalars used as gates change the parity of the gated quantities:
        >>> gate("12x0e + 3x0o + 2x1e + 1x2e")
        12x0e+2x1o+1x2o

        Without anything to gate, all the scalars are activated:
        >>> gate("12x0e + 3x0o")
        12x0e+3x0o
    """
    assert isinstance(input, IrrepsData)

    if even_act is None:
        even_act = jax.nn.gelu
    if odd_act is None:
        odd_act = lambda x: (1 - jnp.exp(-(x**2))) * x
    if even_gate_act is None:
        even_gate_act = jax.nn.sigmoid
    if odd_gate_act is None:
        odd_gate_act = jax.nn.tanh

    return _gate(input, even_act, odd_act, even_gate_act, odd_gate_act)



from functools import partial

import jax
import jax.numpy as jnp

# matrix


def rand_matrix(key, shape):
    r"""random rotation matrix

    Args:
        key: a PRNGKey used as the random key.
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    return angles_to_matrix(*rand_angles(key, shape))


# angles


def identity_angles(shape):
    r"""angles of the identity rotation

    Args:
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return jnp.zeros(shape), jnp.zeros(shape), jnp.zeros(shape)


@partial(jax.jit, static_argnums=(1,), inline=True)
def rand_angles(key, shape):
    r"""random rotation angles

    Args:
        key: a PRNGKey used as the random key.
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    x, y, z = jax.random.uniform(key, (3,) + shape)
    return 2 * jnp.pi * x, jnp.arccos(2 * z - 1), 2 * jnp.pi * y


def compose_angles(a1, b1, c1, a2, b2, c2):
    r"""compose angles

    Computes :math:`(a, b, c)` such that :math:`R(a, b, c) = R(a_1, b_1, c_1) \circ R(a_2, b_2, c_2)`

    Args:
        alpha1 (`jnp.ndarray`): array of shape :math:`(...)`
        beta1 (`jnp.ndarray`): array of shape :math:`(...)`
        gamma1 (`jnp.ndarray`): array of shape :math:`(...)`
        alpha2 (`jnp.ndarray`): array of shape :math:`(...)`
        beta2 (`jnp.ndarray`): array of shape :math:`(...)`
        gamma2 (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`

    """
    a1, b1, c1, a2, b2, c2 = jnp.broadcast_arrays(a1, b1, c1, a2, b2, c2)
    return matrix_to_angles(angles_to_matrix(a1, b1, c1) @ angles_to_matrix(a2, b2, c2))


def inverse_angles(a, b, c):
    r"""angles of the inverse rotation

    Args:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return -c, -b, -a


# quaternions


def identity_quaternion(shape):
    r"""quaternion of identity rotation

    Args:
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 4)`
    """
    q = jnp.zeros(shape, 4)
    return q.at[..., 0].set(1)  # or -1...


@partial(jax.jit, static_argnums=(1,), inline=True)
def rand_quaternion(key, shape):
    r"""generate random quaternion

    Args:
        key: a PRNGKey used as the random key.
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 4)`
    """
    return angles_to_quaternion(*rand_angles(key, shape))


@partial(jax.jit, inline=True)
def compose_quaternion(q1, q2):
    r"""compose two quaternions: :math:`q_1 \circ q_2`

    Args:
        q1 (`jnp.ndarray`): array of shape :math:`(..., 4)`
        q2 (`jnp.ndarray`): array of shape :math:`(..., 4)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 4)`
    """
    q1, q2 = jnp.broadcast_arrays(q1, q2)
    return jnp.stack(
        [
            q1[..., 0] * q2[..., 0] - q1[..., 1] * q2[..., 1] - q1[..., 2] * q2[..., 2] - q1[..., 3] * q2[..., 3],
            q1[..., 1] * q2[..., 0] + q1[..., 0] * q2[..., 1] + q1[..., 2] * q2[..., 3] - q1[..., 3] * q2[..., 2],
            q1[..., 0] * q2[..., 2] - q1[..., 1] * q2[..., 3] + q1[..., 2] * q2[..., 0] + q1[..., 3] * q2[..., 1],
            q1[..., 0] * q2[..., 3] + q1[..., 1] * q2[..., 2] - q1[..., 2] * q2[..., 1] + q1[..., 3] * q2[..., 0],
        ],
        axis=-1,
    )


def inverse_quaternion(q):
    r"""inverse of a quaternion

    Works only for unit quaternions.

    Args:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 4)`
    """
    return q.at[..., 1:].multiply(-1)


# axis-angle


def rand_axis_angle(key, shape):
    r"""generate random rotation as axis-angle

    Args:
        key: a PRNGKey used as the random key.
        shape: a tuple of nonnegative integers representing the result shape.

    Returns:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return angles_to_axis_angle(*rand_angles(key, shape))


def compose_axis_angle(axis1, angle1, axis2, angle2):
    r"""compose :math:`(\vec x_1, \alpha_1)` with :math:`(\vec x_2, \alpha_2)`

    Args:
        axis1 (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle1 (`jnp.ndarray`): array of shape :math:`(...)`
        axis2 (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle2 (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return quaternion_to_axis_angle(
        compose_quaternion(axis_angle_to_quaternion(axis1, angle1), axis_angle_to_quaternion(axis2, angle2))
    )


# conversions


@partial(jax.jit, inline=True)
def matrix_x(angle):
    r"""matrix of rotation around X axis

    Args:
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    c = jnp.cos(angle)
    s = jnp.sin(angle)
    o = jnp.ones_like(angle)
    z = jnp.zeros_like(angle)
    return jnp.stack(
        [
            jnp.stack([o, z, z], axis=-1),
            jnp.stack([z, c, -s], axis=-1),
            jnp.stack([z, s, c], axis=-1),
        ],
        axis=-2,
    )


@partial(jax.jit, inline=True)
def matrix_y(angle):
    r"""matrix of rotation around Y axis

    Args:
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    c = jnp.cos(angle)
    s = jnp.sin(angle)
    o = jnp.ones_like(angle)
    z = jnp.zeros_like(angle)
    return jnp.stack(
        [
            jnp.stack([c, z, s], axis=-1),
            jnp.stack([z, o, z], axis=-1),
            jnp.stack([-s, z, c], axis=-1),
        ],
        axis=-2,
    )


@partial(jax.jit, inline=True)
def matrix_z(angle):
    r"""matrix of rotation around Z axis

    Args:
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    c = jnp.cos(angle)
    s = jnp.sin(angle)
    o = jnp.ones_like(angle)
    z = jnp.zeros_like(angle)
    return jnp.stack([jnp.stack([c, -s, z], axis=-1), jnp.stack([s, c, z], axis=-1), jnp.stack([z, z, o], axis=-1)], axis=-2)


def angles_to_matrix(alpha, beta, gamma):
    r"""conversion from angles to matrix

    Args:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    alpha, beta, gamma = jnp.broadcast_arrays(alpha, beta, gamma)
    return matrix_y(alpha) @ matrix_x(beta) @ matrix_y(gamma)


@partial(jax.jit, inline=True)
def matrix_to_angles(R):
    r"""conversion from matrix to angles

    Args:
        R (`jnp.ndarray`): array of shape :math:`(..., 3, 3)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    # assert jnp.allclose(jnp.linalg.det(R), 1)
    x = R @ jnp.array([0.0, 1.0, 0.0])
    a, b = xyz_to_angles(x)
    R = jnp.swapaxes(angles_to_matrix(a, b, 0.0), -1, -2) @ R
    c = jnp.arctan2(R[..., 0, 2], R[..., 0, 0])
    return a, b, c


@partial(jax.jit, inline=True)
def angles_to_quaternion(alpha, beta, gamma):
    r"""conversion from angles to quaternion

    Args:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`
    """
    alpha, beta, gamma = jnp.broadcast_arrays(alpha, beta, gamma)
    qa = axis_angle_to_quaternion(jnp.array([0.0, 1.0, 0.0]), alpha)
    qb = axis_angle_to_quaternion(jnp.array([1.0, 0.0, 0.0]), beta)
    qc = axis_angle_to_quaternion(jnp.array([0.0, 1.0, 0.0]), gamma)
    return compose_quaternion(qa, compose_quaternion(qb, qc))


def matrix_to_quaternion(R):
    r"""conversion from matrix :math:`R` to quaternion :math:`q`

    Args:
        R (`jnp.ndarray`): array of shape :math:`(..., 3, 3)`

    Returns:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`
    """
    return axis_angle_to_quaternion(*matrix_to_axis_angle(R))


@partial(jax.jit, inline=True)
def axis_angle_to_quaternion(xyz, angle):
    r"""convertion from axis-angle to quaternion

    Args:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`
    """
    xyz, angle = jnp.broadcast_arrays(xyz, angle[..., None])
    xyz = _normalize(xyz)
    c = jnp.cos(angle[..., :1] / 2)
    s = jnp.sin(angle / 2)
    return jnp.concatenate([c, xyz * s], axis=-1)


@partial(jax.jit, inline=True)
def quaternion_to_axis_angle(q):
    r"""convertion from quaternion to axis-angle

    Args:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`

    Returns:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`
    """
    angle = 2 * jnp.arccos(jnp.clip(q[..., 0], -1, 1))
    axis = _normalize(q[..., 1:])
    return axis, angle


def _normalize(x):
    n = jnp.linalg.norm(x, axis=-1, keepdims=True)
    return x / jnp.where(n > 0, n, 1.0)


@partial(jax.jit, inline=True)
def matrix_to_axis_angle(R):
    r"""conversion from matrix to axis-angle

    Args:
        R (`jnp.ndarray`): array of shape :math:`(..., 3, 3)`

    Returns:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`
    """
    # assert jnp.allclose(jnp.linalg.det(R), 1)
    tr = R[..., 0, 0] + R[..., 1, 1] + R[..., 2, 2]
    angle = jnp.arccos(jnp.clip((tr - 1) / 2, -1, 1))
    axis = jnp.stack(
        [
            R[..., 2, 1] - R[..., 1, 2],
            R[..., 0, 2] - R[..., 2, 0],
            R[..., 1, 0] - R[..., 0, 1],
        ],
        axis=-1,
    )
    axis = _normalize(axis)
    return axis, angle


def angles_to_axis_angle(alpha, beta, gamma):
    r"""conversion from angles to axis-angle

    Args:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return matrix_to_axis_angle(angles_to_matrix(alpha, beta, gamma))


@partial(jax.jit, inline=True)
def axis_angle_to_matrix(axis, angle):
    r"""conversion from axis-angle to matrix

    Args:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    axis, angle = jnp.broadcast_arrays(axis, angle[..., None])
    alpha, beta = xyz_to_angles(axis)
    R = angles_to_matrix(alpha, beta, jnp.zeros_like(beta))
    Ry = matrix_y(angle[..., 0])
    return R @ Ry @ jnp.swapaxes(R, -2, -1)


def quaternion_to_matrix(q):
    r"""convertion from quaternion to matrix

    Args:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3, 3)`
    """
    return axis_angle_to_matrix(*quaternion_to_axis_angle(q))


def quaternion_to_angles(q):
    r"""convertion from quaternion to angles

    Args:
        q (`jnp.ndarray`): array of shape :math:`(..., 4)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return matrix_to_angles(quaternion_to_matrix(q))


def axis_angle_to_angles(axis, angle):
    r"""convertion from axis-angle to angles

    Args:
        axis (`jnp.ndarray`): array of shape :math:`(..., 3)`
        angle (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
        gamma (`jnp.ndarray`): array of shape :math:`(...)`
    """
    return matrix_to_angles(axis_angle_to_matrix(axis, angle))


# point on the sphere


@partial(jax.jit, inline=True)
def angles_to_xyz(alpha, beta):
    r"""convert :math:`(\alpha, \beta)` into a point :math:`(x, y, z)` on the sphere

    Args:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`

    Returns:
        `jnp.ndarray`: array of shape :math:`(..., 3)`

    Examples:
        >>> angles_to_xyz(1.7, 0.0) + 0.0
        DeviceArray([0., 1., 0.], dtype=float32, weak_type=True)
    """
    alpha, beta = jnp.broadcast_arrays(alpha, beta)
    x = jnp.sin(beta) * jnp.sin(alpha)
    y = jnp.cos(beta)
    z = jnp.sin(beta) * jnp.cos(alpha)
    return jnp.stack([x, y, z], axis=-1)


@partial(jax.jit, inline=True)
def xyz_to_angles(xyz):
    r"""convert a point :math:`\vec r = (x, y, z)` on the sphere into angles :math:`(\alpha, \beta)`

    .. math::

        \vec r = R(\alpha, \beta, 0) \vec e_y

        \alpha = \arctan(x/z)

        \beta = \arccos(y)

    Args:
        xyz (`jnp.ndarray`): array of shape :math:`(..., 3)`

    Returns:
        alpha (`jnp.ndarray`): array of shape :math:`(...)`
        beta (`jnp.ndarray`): array of shape :math:`(...)`
    """
    xyz = _normalize(xyz)
    xyz = jnp.clip(xyz, -1, 1)

    beta = jnp.arccos(xyz[..., 1])
    alpha = jnp.arctan2(xyz[..., 0], xyz[..., 2])
    return alpha, beta



from functools import wraps

from e3nn_jax import Irreps, IrrepsData


def overload_for_irreps_without_data(irrepsdata_argnums=None, irrepsdata_argnames=None, shape=()):
    def decorator(func):
        from jax._src.api import _infer_argnums_and_argnames

        argnums, argnames = _infer_argnums_and_argnames(func, irrepsdata_argnums, irrepsdata_argnames)

        @wraps(func)
        def wrapper(*args, **kwargs):
            concerned_args = [args[i] for i in argnums if i < len(args)] + [kwargs[k] for k in argnames if k in kwargs]

            if any(isinstance(arg, (Irreps, str)) for arg in concerned_args):
                # assume arguments are Irreps (not IrrepsData)

                args = [IrrepsData.ones(a, shape) if i in argnums else a for i, a in enumerate(args)]
                kwargs = {k: IrrepsData.ones(v, shape) if k in argnames else v for k, v in kwargs.items()}
                output = func(*args, **kwargs)
                if isinstance(output, IrrepsData):
                    return output.irreps
                if isinstance(output, tuple):
                    return tuple(o.irreps if isinstance(o, IrrepsData) else o for o in output)
                raise TypeError(f"{func.__name__} returned {type(output)} which is not supported by `overload_irrep_no_data`.")

            # otherwise, assume arguments are IrrepsData
            return func(*args, **kwargs)

        return wrapper

    return decorator



import haiku as hk
import jax
import jax.numpy as jnp
from e3nn_jax import (
    Irreps,
    IrrepsData,
    FunctionalTensorProduct,
    index_add,
    Linear,
    FullyConnectedTensorProduct,
    MultiLayerPerceptron,
)


class TensorProductMultiLayerPerceptron(hk.Module):
    irreps_in1: Irreps
    irreps_in2: Irreps
    irreps_out: Irreps

    def __init__(self, tp: FunctionalTensorProduct, list_neurons, act):
        super().__init__()

        self.tp = tp
        self.mlp = MultiLayerPerceptron(list_neurons, act)

        self.irreps_in1 = tp.irreps_in1.simplify()
        self.irreps_in2 = tp.irreps_in2.simplify()
        self.irreps_out = tp.irreps_out.simplify()

        assert all(i.has_weight for i in self.tp.instructions)

    def __call__(self, emb, x1: IrrepsData, x2: IrrepsData) -> IrrepsData:
        w = self.mlp(emb)

        w = [
            jnp.einsum(
                "x...,x->...",
                hk.get_parameter(
                    (
                        f"w[{i.i_in1},{i.i_in2},{i.i_out}] "
                        f"{self.tp.irreps_in1[i.i_in1]},{self.tp.irreps_in2[i.i_in2]},{self.tp.irreps_out[i.i_out]}"
                    ),
                    shape=(w.shape[0],) + i.path_shape,
                    init=hk.initializers.RandomNormal(),
                )
                / w.shape[0] ** 0.5,
                w,
            )
            for i in self.tp.instructions
        ]

        x1 = x1.convert(self.irreps_in1)
        x2 = x2.convert(self.irreps_in2)

        return self.tp.left_right(w, x1, x2).convert(self.irreps_out)


def _instructions_uvu(irreps_in1, irreps_in2, out_ir_list):
    irreps_out = []
    instructions = []
    for i1, (mul, ir_in1) in enumerate(irreps_in1):
        for i2, (_, ir_in2) in enumerate(irreps_in2):
            for ir_out in ir_in1 * ir_in2:
                if ir_out in out_ir_list:
                    k = len(irreps_out)
                    irreps_out.append((mul, ir_out))
                    instructions.append((i1, i2, k, "uvu", True))
    irreps_out = Irreps(irreps_out)

    irreps_out, p, _ = irreps_out.sort()
    instructions = [(i_1, i_2, p[i_out], mode, has_weight) for i_1, i_2, i_out, mode, has_weight in instructions]

    return irreps_out, instructions


def _tp_mlp_uvu(emb, input1: IrrepsData, input2: IrrepsData, out_ir_list, *, list_neurons, act) -> IrrepsData:
    irreps_out, instructions = _instructions_uvu(input1.irreps, input2.irreps, out_ir_list)
    tp = FunctionalTensorProduct(input1.irreps, input2.irreps, irreps_out, instructions)
    return TensorProductMultiLayerPerceptron(tp, list_neurons, act)(emb, input1, input2)


def _index_max(i, x, out_dim):
    return jnp.zeros((out_dim,) + x.shape[1:]).at[i].max(x)


class Transformer(hk.Module):
    def __init__(self, irreps_node_output, list_neurons, act, num_heads=1):
        super().__init__()

        self.irreps_node_output = Irreps(irreps_node_output)
        self.list_neurons = list_neurons
        self.act = act
        self.num_heads = num_heads

    def __call__(
        self, edge_src, edge_dst, edge_scalar_attr, edge_weight_cutoff, edge_attr: IrrepsData, node_feat: IrrepsData
    ) -> IrrepsData:
        r"""
        Args:
            edge_src (array of int32): source index of the edges
            edge_dst (array of int32): destination index of the edges
            edge_scalar_attr (array of float): scalar attributes of the edges (typically given by ``soft_one_hot_linspace``)
            edge_weight_cutoff (array of float): cutoff weight for the edges (typically given by ``sus``)
            edge_attr (IrrepsData): attributes of the edges (typically given by ``spherical_harmonics``)
            node_f (IrrepsData): features of the nodes

        Returns:
            IrrepsData: output features of the nodes
        """
        edge_src_feat = jax.tree_map(lambda x: x[edge_src], node_feat)
        edge_dst_feat = jax.tree_map(lambda x: x[edge_dst], node_feat)

        kw = dict(list_neurons=self.list_neurons, act=self.act)
        edge_k = jax.vmap(lambda w, x, y: _tp_mlp_uvu(w, x, y, edge_dst_feat.irreps, **kw))(
            edge_scalar_attr, edge_src_feat, edge_attr
        )  # IrrepData[edge, irreps]
        edge_v = jax.vmap(lambda w, x, y: _tp_mlp_uvu(w, x, y, self.irreps_node_output, **kw))(
            edge_scalar_attr, edge_src_feat, edge_attr
        )  # IrrepData[edge, irreps]

        edge_logit = jax.vmap(FullyConnectedTensorProduct(f"{self.num_heads}x0e"))(
            edge_dst_feat, edge_k
        ).contiguous  # array[edge, head]
        node_logit_max = _index_max(edge_dst, edge_logit, node_feat.shape[0])  # array[node, head]
        exp = edge_weight_cutoff[:, None] * jnp.exp(edge_logit - node_logit_max[edge_dst])  # array[edge, head]
        z = index_add(edge_dst, exp, node_feat.shape[0])  # array[node, head]
        z = jnp.where(z == 0.0, 1.0, z)
        alpha = exp / z[edge_dst]  # array[edge, head]

        edge_v = edge_v.factor_mul_to_last_axis(self.num_heads)  # IrrepsData[edge, head, irreps_out]
        edge_v = edge_v * jnp.sqrt(jax.nn.relu(alpha))  # IrrepsData[edge, head, irreps_out]
        edge_v = edge_v.repeat_mul_by_last_axis()  # IrrepsData[edge, irreps_out]

        node_out = jax.tree_map(lambda x: index_add(edge_dst, x, node_feat.shape[0]), edge_v)  # IrrepsData[node, irreps_out]
        return jax.vmap(Linear(self.irreps_node_output))(node_out)  # IrrepsData[edge, head, irreps_out]



from functools import partial

import haiku as hk
import jax
import jax.numpy as jnp
from e3nn_jax import (
    Irreps,
    IrrepsData,
    FunctionalTensorProduct,
    index_add,
    Linear,
    FullyConnectedTensorProduct,
    MultiLayerPerceptron,
)


class Convolution(hk.Module):
    r"""Equivariant Point Convolution

    Args:
        irreps_node_output : `e3nn.o3.Irreps`
            representation of the output node features

        fc_neurons : list of int
            number of neurons per layers in the fully connected network
            first layer and hidden layers but not the output layer

        num_neighbors : float
            typical number of nodes convolved over
    """

    def __init__(self, irreps_node_output, fc_neurons, num_neighbors, mixing_angle=jnp.pi / 8.0):
        super().__init__()

        self.irreps_node_output = Irreps(irreps_node_output)
        self.fc_neurons = fc_neurons
        self.num_neighbors = num_neighbors
        self.mixing_angle = mixing_angle

    @partial(jax.profiler.annotate_function, name="convolution")
    def __call__(
        self, node_input: IrrepsData, edge_src, edge_dst, edge_attr: IrrepsData, node_attr=None, edge_scalar_attr=None
    ) -> IrrepsData:
        assert isinstance(node_input, IrrepsData)
        assert isinstance(edge_attr, IrrepsData)

        if node_attr is not None:
            tmp = jax.vmap(FullyConnectedTensorProduct(node_input.irreps + self.irreps_node_output))(node_input, node_attr)
        else:
            tmp = jax.vmap(Linear(node_input.irreps + self.irreps_node_output))(node_input)

        node_features, node_self_out = tmp.split([len(node_input.irreps)])

        edge_features = jax.tree_map(lambda x: x[edge_src], node_features)
        del node_features

        ######################################################################################
        irreps_mid = []
        instructions = []
        for i, (mul, ir_in) in enumerate(node_input.irreps):
            for j, (_, ir_edge) in enumerate(edge_attr.irreps):
                for ir_out in ir_in * ir_edge:
                    if ir_out in self.irreps_node_output or ir_out.is_scalar():
                        k = len(irreps_mid)
                        irreps_mid.append((mul, ir_out))
                        instructions.append((i, j, k, "uvu", True))
        irreps_mid = Irreps(irreps_mid)

        assert irreps_mid.dim > 0, (
            f"irreps_node_input={node_input.irreps} "
            f"time irreps_edge_attr={edge_attr.irreps} "
            f"produces nothing in irreps_node_output={self.irreps_node_output}"
        )

        irreps_mid, p, _ = irreps_mid.sort()
        instructions = [(i_1, i_2, p[i_out], mode, train) for i_1, i_2, i_out, mode, train in instructions]

        tp = FunctionalTensorProduct(
            node_input.irreps,
            edge_attr.irreps,
            irreps_mid,
            instructions,
        )

        if self.fc_neurons:
            weight = MultiLayerPerceptron(self.fc_neurons, jax.nn.gelu)(edge_scalar_attr)

            weight = [
                jnp.einsum(
                    "x...,ex->e...",
                    hk.get_parameter(
                        (
                            f"w[{ins.i_in1},{ins.i_in2},{ins.i_out}] "
                            f"{tp.irreps_in1[ins.i_in1]},{tp.irreps_in2[ins.i_in2]},{tp.irreps_out[ins.i_out]}"
                        ),
                        shape=(weight.shape[1],) + ins.path_shape,
                        init=hk.initializers.RandomNormal(),
                    )
                    / weight.shape[1] ** 0.5,
                    weight,
                )
                for ins in tp.instructions
            ]

            edge_features: IrrepsData = jax.vmap(tp.left_right, (0, 0, 0), 0)(weight, edge_features, edge_attr)
        else:
            weight = [
                hk.get_parameter(
                    (
                        f"w[{ins.i_in1},{ins.i_in2},{ins.i_out}] "
                        f"{tp.irreps_in1[ins.i_in1]},{tp.irreps_in2[ins.i_in2]},{tp.irreps_out[ins.i_out]}"
                    ),
                    shape=ins.path_shape,
                    init=hk.initializers.RandomNormal(),
                )
                for ins in tp.instructions
            ]
            edge_features: IrrepsData = jax.vmap(tp.left_right, (None, 0, 0), 0)(weight, edge_features, edge_attr)

        edge_features = edge_features.remove_nones().simplify()
        ######################################################################################

        node_features = jax.tree_map(lambda x: index_add(edge_dst, x, out_dim=node_input.shape[0]), edge_features)
        node_features = node_features / self.num_neighbors**0.5

        ######################################################################################

        if node_attr is not None:
            node_conv_out = jax.vmap(FullyConnectedTensorProduct(self.irreps_node_output))(node_features, node_attr)
        else:
            node_conv_out = jax.vmap(Linear(self.irreps_node_output))(node_features)

        ######################################################################################

        with jax.ensure_compile_time_eval():
            c = jnp.cos(self.mixing_angle)
            s = jnp.sin(self.mixing_angle)

        return c * node_self_out + s * node_conv_out



"""Defines the functional tensor product."""

import itertools
from functools import lru_cache, partial
from math import sqrt
import collections
from typing import Any, List, Dict, Optional

import jax
import jax.numpy as jnp

from e3nn_jax import Irreps, IrrepsData, clebsch_gordan, Instruction
from e3nn_jax.util import prod

from ._einsum import einsum as opt_einsum


def _sum_tensors(xs, shape, empty_return_none=False):
    xs = [x for x in xs if x is not None]
    if len(xs) > 0:
        out = xs[0].reshape(shape)
        for x in xs[1:]:
            out = out + x.reshape(shape)
        return out
    if empty_return_none:
        return None
    return jnp.zeros(shape)


def _flat_concatenate(xs):
    if any(x is None for x in xs):
        return None
    if len(xs) > 0:
        return jnp.concatenate([x.flatten() for x in xs])
    return jnp.zeros((0,))


def _compute_element_path_normalization_factors(
    instructions: List[Instruction],
    first_input_variance: List[float],
    second_input_variance: List[float],
) -> Dict[Instruction, float]:
    """Returns a dictionary with keys as the Instructions and values as the corresponding path normalization factor
    for 'element' path normalization."""
    path_normalization_sums = collections.defaultdict(lambda: 0.0)
    for instruction in instructions:
        path_normalization_sums[instruction.i_out] += (
            first_input_variance[instruction.i_in1] * second_input_variance[instruction.i_in2] * instruction.num_elements
        )

    return {instruction: path_normalization_sums[instruction.i_out] for instruction in instructions}


def _compute_standard_path_normalization_factors(
    instructions: List[Instruction],
    first_input_variance: List[float],
    second_input_variance: List[float],
) -> Dict[Instruction, float]:
    """Returns a dictionary with keys as the Instructions and values as the corresponding path normalization factor
    for 'path' path normalization."""
    path_normalization_counts = collections.defaultdict(lambda: 0.0)
    for instruction in instructions:
        path_normalization_counts[instruction.i_out] += 1

    return {
        instruction: first_input_variance[instruction.i_in1]
        * second_input_variance[instruction.i_in2]
        * instruction.num_elements
        * path_normalization_counts[instruction.i_out]
        for instruction in instructions
    }


def normalize_instruction_path_weights(
    instructions: List[Instruction],
    first_input_irreps: Irreps,
    second_input_irreps: Irreps,
    output_irreps: Irreps,
    first_input_variance: List[float],
    second_input_variance: List[float],
    output_variance: List[float],
    irrep_normalization: str,
    path_normalization: str,
) -> List[Instruction]:
    """Returns instructions with normalized path weights."""
    # Precompute normalization factors.
    if path_normalization == "element":
        path_normalization_factors = _compute_element_path_normalization_factors(
            instructions, first_input_variance, second_input_variance
        )
    elif path_normalization == "path":
        path_normalization_factors = _compute_standard_path_normalization_factors(
            instructions, first_input_variance, second_input_variance
        )
    else:
        raise ValueError(f"Unsupported path normalization: {path_normalization}.")

    def compute_normalized_path_weight_fn(instruction: Instruction) -> float:
        return compute_normalized_path_weight(
            instruction,
            first_input_irreps,
            second_input_irreps,
            output_irreps,
            output_variance,
            irrep_normalization,
            path_normalization_factors,
        )

    return [instruction.replace(path_weight=compute_normalized_path_weight_fn(instruction)) for instruction in instructions]


def compute_normalized_path_weight(
    instruction: Instruction,
    first_input_irreps: Irreps,
    second_input_irreps: Irreps,
    output_irreps: Irreps,
    output_variance: List[float],
    irrep_normalization: str,
    path_normalization_factors: Dict[Instruction, float],
) -> float:
    """Computes normalized path weight for a single instructions, with precomputed path normalization factors."""

    if irrep_normalization not in ["component", "norm", "none"]:
        raise ValueError(f"Unsupported irrep normalization: {irrep_normalization}.")

    mul_ir_in1 = first_input_irreps[instruction.i_in1]
    mul_ir_in2 = second_input_irreps[instruction.i_in2]
    mul_ir_out = output_irreps[instruction.i_out]

    assert mul_ir_in1.ir.p * mul_ir_in2.ir.p == mul_ir_out.ir.p
    assert abs(mul_ir_in1.ir.l - mul_ir_in2.ir.l) <= mul_ir_out.ir.l <= mul_ir_in1.ir.l + mul_ir_in2.ir.l

    if irrep_normalization == "component":
        alpha = mul_ir_out.ir.dim
    if irrep_normalization == "norm":
        alpha = mul_ir_in1.ir.dim * mul_ir_in2.ir.dim
    if irrep_normalization == "none":
        alpha = 1

    path_normalization_factor = path_normalization_factors[instruction]
    if path_normalization_factor > 0.0:
        alpha /= path_normalization_factor

    alpha *= output_variance[instruction.i_out]
    alpha *= instruction.path_weight
    return sqrt(alpha)


class FunctionalTensorProduct:
    r"""Tensor product of two tensors.

    Args:
        irreps_in1: :class:`~e3nn_jax.Irreps` of the first tensor.
        irreps_in2: :class:`~e3nn_jax.Irreps` of the second tensor.
        irreps_out: :class:`~e3nn_jax.Irreps` of the output tensor.
        instructions: List of instructions.
            ``[(i_in1, i_in2, i_out, connection_mode, has_weight, (path_weight)), ...]``
            - i_in1, i_in2, i_out are indices of the irreps_in1, irreps_in2, irreps_out.
            - connection_mode is one of ``uvw``, ``uvu``, ``uvv``, ``uuw``, ``uuu``, ``uvuv``
            - has_weight is a boolean indicating whether the instruction has a weight.
            - path_weight (optional) is the weight of the path.

        in1_var: Variance of the first tensor.
        in2_var: Variance of the second tensor.
        out_var: Variance of the output tensor.
        irrep_normalization: Normalization of the tensors. `component` or `norm`.
        path_normalization: Normalization of the paths. `element` or `path`.
    """
    irreps_in1: Irreps
    irreps_in2: Irreps
    irreps_out: Irreps
    instructions: List[Instruction]
    output_mask: jnp.ndarray

    def __init__(
        self,
        irreps_in1: Any,
        irreps_in2: Any,
        irreps_out: Any,
        instructions: List[Any],
        in1_var: Optional[List[float]] = None,
        in2_var: Optional[List[float]] = None,
        out_var: Optional[List[float]] = None,
        irrep_normalization: str = "component",
        path_normalization: str = "element",
    ):
        self.irreps_in1 = Irreps(irreps_in1)
        self.irreps_in2 = Irreps(irreps_in2)
        self.irreps_out = Irreps(irreps_out)
        del irreps_in1, irreps_in2, irreps_out

        instructions = [x if len(x) == 6 else x + (1.0,) for x in instructions]
        instructions = [
            Instruction(
                i_in1,
                i_in2,
                i_out,
                connection_mode,
                has_weight,
                path_weight,
                self.irreps_in1[i_in1].mul,
                self.irreps_in2[i_in2].mul,
                self.irreps_out[i_out].mul,
            )
            for i_in1, i_in2, i_out, connection_mode, has_weight, path_weight in instructions
        ]

        if in1_var is None:
            in1_var = [1.0 for _ in self.irreps_in1]

        if in2_var is None:
            in2_var = [1.0 for _ in self.irreps_in2]

        if out_var is None:
            out_var = [1.0 for _ in self.irreps_out]

        self.instructions = normalize_instruction_path_weights(
            instructions,
            self.irreps_in1,
            self.irreps_in2,
            self.irreps_out,
            in1_var,
            in2_var,
            out_var,
            irrep_normalization,
            path_normalization,
        )

        with jax.ensure_compile_time_eval():
            if self.irreps_out.dim > 0:
                self.output_mask = jnp.concatenate(
                    [
                        jnp.ones(mul_ir.dim)
                        if any(
                            (ins.i_out == i_out) and (ins.path_weight != 0) and (0 not in ins.path_shape)
                            for ins in self.instructions
                        )
                        else jnp.zeros(mul_ir.dim)
                        for i_out, mul_ir in enumerate(self.irreps_out)
                    ]
                )
            else:
                self.output_mask = jnp.ones(0)

    def left_right(
        self,
        weights: List[jnp.ndarray],
        input1: IrrepsData,
        input2: IrrepsData = None,
        *,
        specialized_code=False,
        optimize_einsums=True,
        custom_einsum_vjp=False,
        fuse_all=False,
    ) -> IrrepsData:
        r"""Compute the tensor product of two input tensors.

        Args:
            weights (array or list of arrays): The weights of the tensor product.
            input1 (IrrepsData): The first input tensor.
            input2 (IrrepsData): The second input tensor.
            specialized_code (bool): If True, use the specialized code for the
                tensor product.
            optimize_einsums (bool): If True, optimize the einsum code.
            custom_einsum_vjp (bool): If True, use the custom vjp for the einsum
                code.
            fuse_all (bool): If True, fuse all the einsums.

        Returns:
            `IrrepsData`: The output tensor.
        """
        if input2 is None:
            weights, input1, input2 = [], weights, input1

        input1 = IrrepsData.new(self.irreps_in1, input1)
        input2 = IrrepsData.new(self.irreps_in2, input2)

        return _left_right(
            self,
            weights,
            input1,
            input2,
            specialized_code=specialized_code,
            optimize_einsums=optimize_einsums,
            custom_einsum_vjp=custom_einsum_vjp,
            fuse_all=fuse_all,
        )

    def right(
        self,
        weights: List[jnp.ndarray],
        input2: IrrepsData = None,
        *,
        optimize_einsums=False,
        custom_einsum_vjp=False,
    ) -> jnp.ndarray:
        r"""Compute the right contraction of the tensor product.

        Args:
            weights (array or list of arrays): The weights of the tensor product.
            input2 (IrrepsData): The second input tensor.
            optimize_einsums (bool): If True, optimize the einsum code.
            custom_einsum_vjp (bool): If True, use the custom vjp for the einsum code.

        Returns:
            A matrix of shape ``(irreps_in1.dim, irreps_out.dim)``.
        """
        if input2 is None:
            weights, input2 = [], weights

        input2 = IrrepsData.new(self.irreps_in2, input2)
        return _right(
            self,
            weights,
            input2,
            optimize_einsums=optimize_einsums,
            custom_einsum_vjp=custom_einsum_vjp,
        )

    def __repr__(self):
        npath = sum(prod(i.path_shape) for i in self.instructions)
        nweight = sum(prod(i.path_shape) for i in self.instructions if i.has_weight)
        return (
            f"{self.__class__.__name__}"
            f"({self.irreps_in1.simplify()} x {self.irreps_in2.simplify()} "
            f"-> {self.irreps_out.simplify()} | {npath} paths | {nweight} weights)"
        )


@partial(
    jax.jit, static_argnums=(0,), static_argnames=("specialized_code", "optimize_einsums", "custom_einsum_vjp", "fuse_all")
)
@partial(jax.profiler.annotate_function, name="TensorProduct.left_right")
def _left_right(
    self: FunctionalTensorProduct,
    weights,
    input1,
    input2,
    *,
    specialized_code=False,
    optimize_einsums=True,
    custom_einsum_vjp=False,
    fuse_all=False,
):

    # = Short-circut for zero dimensional =
    if self.irreps_in1.dim == 0 or self.irreps_in2.dim == 0 or self.irreps_out.dim == 0:
        return IrrepsData.zeros(self.irreps_out, ())

    if custom_einsum_vjp:
        assert optimize_einsums
        einsum = opt_einsum
    else:
        einsum = partial(jnp.einsum, optimize="optimal" if optimize_einsums else "greedy")

    if isinstance(weights, list):
        assert len(weights) == len([ins for ins in self.instructions if ins.has_weight]), (
            len(weights),
            len([ins for ins in self.instructions if ins.has_weight]),
        )
        weights_flat = _flat_concatenate(weights)
        weights_list = weights
    else:
        weights_flat = weights
        weights_list = []
        i = 0
        for ins in self.instructions:
            if ins.has_weight:
                n = prod(ins.path_shape)
                weights_list.append(weights[i : i + n].reshape(ins.path_shape))
                i += n
        assert i == weights.size
    del weights

    assert len(input1.shape) == 0
    assert len(input2.shape) == 0

    if fuse_all:
        with jax.ensure_compile_time_eval():
            num_path = weights_flat.size
            has_path_with_no_weights = any(not ins.has_weight for ins in self.instructions)
            i = 0

            if has_path_with_no_weights:
                num_path += 1
                i += 1

            big_w3j = jnp.zeros(
                (
                    num_path,
                    self.irreps_in1.dim,
                    self.irreps_in2.dim,
                    self.irreps_out.dim,
                )
            )
            for ins in self.instructions:
                mul_ir_in1 = self.irreps_in1[ins.i_in1]
                mul_ir_in2 = self.irreps_in2[ins.i_in2]
                mul_ir_out = self.irreps_out[ins.i_out]
                m1, m2, mo = mul_ir_in1.mul, mul_ir_in2.mul, mul_ir_out.mul
                d1, d2, do = mul_ir_in1.ir.dim, mul_ir_in2.ir.dim, mul_ir_out.ir.dim
                s1 = self.irreps_in1[: ins.i_in1].dim
                s2 = self.irreps_in2[: ins.i_in2].dim
                so = self.irreps_out[: ins.i_out].dim

                w3j = clebsch_gordan(mul_ir_in1.ir.l, mul_ir_in2.ir.l, mul_ir_out.ir.l)

                def set_w3j(i, u, v, w):
                    return big_w3j.at[
                        i,
                        s1 + u * d1 : s1 + (u + 1) * d1,
                        s2 + v * d2 : s2 + (v + 1) * d2,
                        so + w * do : so + (w + 1) * do,
                    ].add(ins.path_weight * w3j)

                if ins.connection_mode == "uvw":
                    assert ins.has_weight
                    for u, v, w in itertools.product(range(m1), range(m2), range(mo)):
                        big_w3j = set_w3j(i, u, v, w)
                        i += 1
                elif ins.connection_mode == "uvu":
                    assert ins.has_weight
                    for u, v in itertools.product(range(m1), range(m2)):
                        big_w3j = set_w3j(i, u, v, u)
                        i += 1
                elif ins.connection_mode == "uvv":
                    assert ins.has_weight
                    for u, v in itertools.product(range(m1), range(m2)):
                        big_w3j = set_w3j(i, u, v, v)
                        i += 1
                elif ins.connection_mode == "uuu":
                    for u in range(m1):
                        if ins.has_weight:
                            big_w3j = set_w3j(i, u, u, u)
                            i += 1
                        else:
                            big_w3j = set_w3j(0, u, u, u)
                else:
                    assert False

        if has_path_with_no_weights and big_w3j.shape[0] == 1:
            big_w3j = big_w3j.reshape(big_w3j.shape[1:])
            out = einsum("ijk,i,j->k", big_w3j, input1.contiguous, input2.contiguous)
        else:
            if has_path_with_no_weights:
                weights_flat = jnp.concatenate([jnp.ones((1,)), weights_flat])

            out = einsum(
                "p,pijk,i,j->k",
                weights_flat,
                big_w3j,
                input1.contiguous,
                input2.contiguous,
            )
        return IrrepsData.from_contiguous(self.irreps_out, out)

    @lru_cache(maxsize=None)
    def multiply(in1, in2, mode):
        if mode == "uv":
            return einsum("ui,vj->uvij", input1.list[in1], input2.list[in2])
        if mode == "uu":
            return einsum("ui,uj->uij", input1.list[in1], input2.list[in2])

    weight_index = 0

    out_list = []

    for ins in self.instructions:
        mul_ir_in1 = self.irreps_in1[ins.i_in1]
        mul_ir_in2 = self.irreps_in2[ins.i_in2]
        mul_ir_out = self.irreps_out[ins.i_out]

        if ins.has_weight:
            w = weights_list[weight_index]
            assert w.shape == ins.path_shape
            weight_index += 1

        if mul_ir_in1.dim == 0 or mul_ir_in2.dim == 0 or mul_ir_out.dim == 0:
            # TODO verify that there is no need for
            # out_list += [None]
            continue

        x1 = input1.list[ins.i_in1]
        x2 = input2.list[ins.i_in2]
        if x1 is None or x2 is None:
            out_list += [None]
            continue

        xx = multiply(ins.i_in1, ins.i_in2, ins.connection_mode[:2])

        with jax.ensure_compile_time_eval():
            w3j = clebsch_gordan(mul_ir_in1.ir.l, mul_ir_in2.ir.l, mul_ir_out.ir.l)
            w3j = ins.path_weight * w3j

        if ins.connection_mode == "uvw":
            assert ins.has_weight
            if specialized_code and (
                mul_ir_in1.ir.l,
                mul_ir_in2.ir.l,
                mul_ir_out.ir.l,
            ) == (0, 0, 0):
                out = ins.path_weight * einsum("uvw,uv->w", w, xx.reshape(mul_ir_in1.dim, mul_ir_in2.dim))
            elif specialized_code and mul_ir_in1.ir.l == 0:
                out = ins.path_weight * einsum("uvw,u,vj->wj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
            elif specialized_code and mul_ir_in2.ir.l == 0:
                out = ins.path_weight * einsum("uvw,ui,v->wi", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
            elif specialized_code and mul_ir_out.ir.l == 0:
                out = ins.path_weight * einsum("uvw,ui,vi->w", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
            else:
                out = einsum("uvw,ijk,uvij->wk", w, w3j, xx)
        if ins.connection_mode == "uvu":
            assert mul_ir_in1.mul == mul_ir_out.mul
            if ins.has_weight:
                if specialized_code and (
                    mul_ir_in1.ir.l,
                    mul_ir_in2.ir.l,
                    mul_ir_out.ir.l,
                ) == (0, 0, 0):
                    out = ins.path_weight * einsum(
                        "uv,u,v->u",
                        w,
                        x1.reshape(mul_ir_in1.dim),
                        x2.reshape(mul_ir_in2.dim),
                    )
                elif specialized_code and mul_ir_in1.ir.l == 0:
                    out = ins.path_weight * einsum("uv,u,vj->uj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
                elif specialized_code and mul_ir_in2.ir.l == 0:
                    out = ins.path_weight * einsum("uv,ui,v->ui", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
                elif specialized_code and mul_ir_out.ir.l == 0:
                    out = ins.path_weight * einsum("uv,ui,vi->u", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
                else:
                    out = einsum("uv,ijk,uvij->uk", w, w3j, xx)
            else:
                # not so useful operation because v is summed
                out = einsum("ijk,uvij->uk", w3j, xx)
        if ins.connection_mode == "uvv":
            assert mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                if specialized_code and (
                    mul_ir_in1.ir.l,
                    mul_ir_in2.ir.l,
                    mul_ir_out.ir.l,
                ) == (0, 0, 0):
                    out = ins.path_weight * einsum(
                        "uv,u,v->v",
                        w,
                        x1.reshape(mul_ir_in1.dim),
                        x2.reshape(mul_ir_in2.dim),
                    )
                elif specialized_code and mul_ir_in1.ir.l == 0:
                    out = ins.path_weight * einsum("uv,u,vj->vj", w, x1.reshape(mul_ir_in1.dim), x2) / sqrt(mul_ir_out.ir.dim)
                elif specialized_code and mul_ir_in2.ir.l == 0:
                    out = ins.path_weight * einsum("uv,ui,v->vi", w, x1, x2.reshape(mul_ir_in2.dim)) / sqrt(mul_ir_out.ir.dim)
                elif specialized_code and mul_ir_out.ir.l == 0:
                    out = ins.path_weight * einsum("uv,ui,vi->v", w, x1, x2) / sqrt(mul_ir_in1.ir.dim)
                else:
                    out = einsum("uv,ijk,uvij->vk", w, w3j, xx)
            else:
                # not so useful operation because u is summed
                out = einsum("ijk,uvij->vk", w3j, xx)
        if ins.connection_mode == "uuw":
            assert mul_ir_in1.mul == mul_ir_in2.mul
            if ins.has_weight:
                out = einsum("uw,ijk,uij->wk", w, w3j, xx)
            else:
                # equivalent to tp(x, y, 'uuu').sum('u')
                assert mul_ir_out.mul == 1
                out = einsum("ijk,uij->k", w3j, xx)
        if ins.connection_mode == "uuu":
            assert mul_ir_in1.mul == mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("u,ijk,uij->uk", w, w3j, xx)
            else:
                out = einsum("ijk,uij->uk", w3j, xx)
        if ins.connection_mode == "uvuv":
            assert mul_ir_in1.mul * mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("uv,ijk,uvij->uvk", w, w3j, xx)
            else:
                out = einsum("ijk,uvij->uvk", w3j, xx)
        if ins.connection_mode == "uvu<v":
            assert mul_ir_in1.mul == mul_ir_in2.mul
            assert mul_ir_in1.mul * (mul_ir_in1.mul - 1) // 2 == mul_ir_out.mul
            i = jnp.triu_indices(mul_ir_in1.mul, 1)
            xx = xx[i[0], i[1]]  # uvij -> wij
            if ins.has_weight:
                out = einsum("w,ijk,wij->wk", w, w3j, xx)
            else:
                out = einsum("ijk,wij->wk", w3j, xx)
        if ins.connection_mode == "u<vw":
            assert mul_ir_in1.mul == mul_ir_in2.mul
            assert ins.has_weight
            i = jnp.triu_indices(mul_ir_in1.mul, 1)
            xx = multiply(ins.i_in1, ins.i_in2, "uv")
            xx = xx[i[0], i[1]]  # uvij -> qij
            out = einsum("qw,ijk,qij->wk", w, w3j, xx)

        out_list += [out]

    out = [
        _sum_tensors(
            [out for ins, out in zip(self.instructions, out_list) if ins.i_out == i_out],
            shape=(mul_ir_out.mul, mul_ir_out.ir.dim),
            empty_return_none=True,
        )
        for i_out, mul_ir_out in enumerate(self.irreps_out)
    ]
    return IrrepsData.from_list(self.irreps_out, out, ())


@partial(jax.jit, static_argnums=(0,), static_argnames=("optimize_einsums", "custom_einsum_vjp"))
@partial(jax.profiler.annotate_function, name="TensorProduct.right")
def _right(
    self: FunctionalTensorProduct,
    weights,
    input2,
    *,
    optimize_einsums=False,
    custom_einsum_vjp=False,
):
    # = Short-circut for zero dimensional =
    if self.irreps_in1.dim == 0 or self.irreps_in2.dim == 0 or self.irreps_out.dim == 0:
        return jnp.zeros(
            (
                self.irreps_in1.dim,
                self.irreps_out.dim,
            )
        )

    if custom_einsum_vjp:
        assert optimize_einsums
        einsum = opt_einsum
    else:
        einsum = partial(jnp.einsum, optimize="optimal" if optimize_einsums else "greedy")

    weight_index = 0

    out_list = []

    for ins in self.instructions:
        mul_ir_in1 = self.irreps_in1[ins.i_in1]
        mul_ir_in2 = self.irreps_in2[ins.i_in2]
        mul_ir_out = self.irreps_out[ins.i_out]

        x2 = input2.list[ins.i_in2]

        if ins.has_weight:
            w = weights[weight_index]
            assert w.shape == ins.path_shape, (
                w.shape,
                ins.path_shape,
                weight_index,
                ins,
            )
            weight_index += 1

        if mul_ir_in1.dim == 0 or mul_ir_in2.dim == 0 or mul_ir_out.dim == 0:
            # TODO add tests for this case
            out_list += [jnp.zeros((mul_ir_in1.dim, mul_ir_out.dim))]
            continue

        with jax.ensure_compile_time_eval():
            w3j = clebsch_gordan(mul_ir_in1.ir.l, mul_ir_in2.ir.l, mul_ir_out.ir.l)

        if ins.connection_mode == "uvw":
            assert ins.has_weight
            out = einsum("uvw,ijk,vj->uiwk", w, w3j, x2)
        if ins.connection_mode == "uvu":
            assert mul_ir_in1.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("uv,ijk,vj,uw->uiwk", w, w3j, x2, jnp.eye(mul_ir_in1.mul))
            else:
                # not so useful operation because v is summed
                out = einsum("ijk,vj,uw->uiwk", w3j, x2, jnp.eye(mul_ir_in1.mul))
        if ins.connection_mode == "uvv":
            assert mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("uv,ijk,vj->uivk", w, w3j, x2)
            else:
                # not so useful operation because u is summed
                out = einsum("ijk,vj,u->uivk", w3j, x2, jnp.ones((mul_ir_in1.mul,)))
        if ins.connection_mode == "uuw":
            assert mul_ir_in1.mul == mul_ir_in2.mul
            if ins.has_weight:
                out = einsum("uw,ijk,uj->uiwk", w, w3j, x2)
            else:
                # equivalent to tp(x, y, 'uuu').sum('u')
                assert mul_ir_out.mul == 1
                out = einsum("ijk,uj->uik", w3j, x2)
        if ins.connection_mode == "uuu":
            assert mul_ir_in1.mul == mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("u,ijk,uj,uw->uiwk", w, w3j, x2, jnp.eye(mul_ir_in1.mul))
            else:
                out = einsum("ijk,uj,uw->uiwk", w3j, x2, jnp.eye(mul_ir_in1.mul))
        if ins.connection_mode == "uvuv":
            assert mul_ir_in1.mul * mul_ir_in2.mul == mul_ir_out.mul
            if ins.has_weight:
                out = einsum("uv,ijk,vj,uw->uiwvk", w, w3j, x2, jnp.eye(mul_ir_in1.mul))
            else:
                out = einsum("ijk,vj,uw->uiwvk", w3j, x2, jnp.eye(mul_ir_in1.mul))

        out = ins.path_weight * out

        out_list += [out.reshape(mul_ir_in1.dim, mul_ir_out.dim)]

    return jnp.concatenate(
        [
            jnp.concatenate(
                [
                    _sum_tensors(
                        [out for ins, out in zip(self.instructions, out_list) if (ins.i_in1, ins.i_out) == (i_in1, i_out)],
                        shape=(mul_ir_in1.dim, mul_ir_out.dim),
                    )
                    for i_out, mul_ir_out in enumerate(self.irreps_out)
                    if mul_ir_out.mul > 0
                ],
                axis=1,
            )
            for i_in1, mul_ir_in1 in enumerate(self.irreps_in1)
            if mul_ir_in1.mul > 0
        ],
        axis=0,
    )



from typing import Any, List, Optional

import haiku as hk
import jax
import jax.numpy as jnp

from e3nn_jax import FunctionalTensorProduct, Irrep, Irreps, IrrepsData
from e3nn_jax.util.decorators import overload_for_irreps_without_data


def naive_broadcast_decorator(func):
    def wrapper(*args):
        shape = jnp.broadcast_shapes(*(arg.shape for arg in args))
        args = [arg.broadcast_to(shape) for arg in args]
        f = func
        for _ in range(len(shape)):
            f = jax.vmap(f)
        return f(*args)

    return wrapper


def FunctionalFullyConnectedTensorProduct(
    irreps_in1: Any,
    irreps_in2: Any,
    irreps_out: Any,
    in1_var: Optional[List[float]] = None,
    in2_var: Optional[List[float]] = None,
    out_var: Optional[List[float]] = None,
    irrep_normalization: str = "component",
    path_normalization: str = "element",
):
    irreps_in1 = Irreps(irreps_in1)
    irreps_in2 = Irreps(irreps_in2)
    irreps_out = Irreps(irreps_out)

    instructions = [
        (i_1, i_2, i_out, "uvw", True)
        for i_1, (_, ir_1) in enumerate(irreps_in1)
        for i_2, (_, ir_2) in enumerate(irreps_in2)
        for i_out, (_, ir_out) in enumerate(irreps_out)
        if ir_out in ir_1 * ir_2
    ]
    return FunctionalTensorProduct(
        irreps_in1, irreps_in2, irreps_out, instructions, in1_var, in2_var, out_var, irrep_normalization, path_normalization
    )


class FullyConnectedTensorProduct(hk.Module):
    def __init__(self, irreps_out, *, irreps_in1=None, irreps_in2=None):
        super().__init__()

        self.irreps_out = Irreps(irreps_out)
        self.irreps_in1 = Irreps(irreps_in1) if irreps_in1 is not None else None
        self.irreps_in2 = Irreps(irreps_in2) if irreps_in2 is not None else None

    def __call__(self, x1: IrrepsData, x2: IrrepsData) -> IrrepsData:
        if self.irreps_in1 is not None:
            x1 = IrrepsData.new(self.irreps_in1, x1)
        if self.irreps_in2 is not None:
            x2 = IrrepsData.new(self.irreps_in2, x2)

        x1 = x1.remove_nones().simplify()
        x2 = x2.remove_nones().simplify()

        tp = FunctionalFullyConnectedTensorProduct(x1.irreps, x2.irreps, self.irreps_out.simplify())
        ws = [
            hk.get_parameter(
                (
                    f"w[{ins.i_in1},{ins.i_in2},{ins.i_out}] "
                    f"{tp.irreps_in1[ins.i_in1]},{tp.irreps_in2[ins.i_in2]},{tp.irreps_out[ins.i_out]}"
                ),
                shape=ins.path_shape,
                init=hk.initializers.RandomNormal(),
            )
            for ins in tp.instructions
        ]
        f = naive_broadcast_decorator(lambda x1, x2: tp.left_right(ws, x1, x2))
        output = f(x1, x2)
        return output.convert(self.irreps_out)


@overload_for_irreps_without_data((0, 1))
def full_tensor_product(
    input1: IrrepsData,
    input2: IrrepsData,
    filter_ir_out=None,
    irrep_normalization: str = "component",
):
    if filter_ir_out is not None:
        filter_ir_out = [Irrep(ir) for ir in filter_ir_out]

    irreps_out = []
    instructions = []
    for i_1, (mul_1, ir_1) in enumerate(input1.irreps):
        for i_2, (mul_2, ir_2) in enumerate(input2.irreps):
            for ir_out in ir_1 * ir_2:

                if filter_ir_out is not None and ir_out not in filter_ir_out:
                    continue

                i_out = len(irreps_out)
                irreps_out.append((mul_1 * mul_2, ir_out))
                instructions += [(i_1, i_2, i_out, "uvuv", False)]

    irreps_out = Irreps(irreps_out)
    irreps_out, p, _ = irreps_out.sort()

    instructions = [(i_1, i_2, p[i_out], mode, train) for i_1, i_2, i_out, mode, train in instructions]

    tp = FunctionalTensorProduct(
        input1.irreps, input2.irreps, irreps_out, instructions, irrep_normalization=irrep_normalization
    )

    return naive_broadcast_decorator(tp.left_right)(input1, input2)


@overload_for_irreps_without_data((0, 1))
def elementwise_tensor_product(
    input1: IrrepsData,
    input2: IrrepsData,
    filter_ir_out=None,
    irrep_normalization: str = "component",
    path_normalization: str = "element",
) -> IrrepsData:
    if filter_ir_out is not None:
        filter_ir_out = [Irrep(ir) for ir in filter_ir_out]

    assert input1.irreps.num_irreps == input2.irreps.num_irreps

    irreps_in1 = list(input1.irreps)
    irreps_in2 = list(input2.irreps)

    i = 0
    while i < len(irreps_in1):
        mul_1, ir_1 = irreps_in1[i]
        mul_2, ir_2 = irreps_in2[i]

        if mul_1 < mul_2:
            irreps_in2[i] = (mul_1, ir_2)
            irreps_in2.insert(i + 1, (mul_2 - mul_1, ir_2))

        if mul_2 < mul_1:
            irreps_in1[i] = (mul_2, ir_1)
            irreps_in1.insert(i + 1, (mul_1 - mul_2, ir_1))
        i += 1

    input1 = input1.convert(irreps_in1)
    input2 = input2.convert(irreps_in2)

    irreps_out = []
    instructions = []
    for i, ((mul, ir_1), (mul_2, ir_2)) in enumerate(zip(irreps_in1, irreps_in2)):
        assert mul == mul_2
        for ir in ir_1 * ir_2:

            if filter_ir_out is not None and ir not in filter_ir_out:
                continue

            i_out = len(irreps_out)
            irreps_out.append((mul, ir))
            instructions += [(i, i, i_out, "uuu", False)]

    tp = FunctionalTensorProduct(
        irreps_in1,
        irreps_in2,
        irreps_out,
        instructions,
        irrep_normalization=irrep_normalization,
        path_normalization=path_normalization,
    )

    return naive_broadcast_decorator(tp.left_right)(input1, input2)


def FunctionalTensorSquare(irreps_in: Irreps, irreps_out: Irreps, irrep_normalization: str = None, **kwargs):
    if irrep_normalization is None:
        irrep_normalization = "component"

    assert irrep_normalization in ["component", "norm", "none"]

    irreps_in = Irreps(irreps_in)
    irreps_out = Irreps(irreps_out)

    instructions = []
    for i_1, (mul_1, ir_1) in enumerate(irreps_in):
        for i_2, (mul_2, ir_2) in enumerate(irreps_in):
            for i_out, (mul_out, ir_out) in enumerate(irreps_out):
                if ir_out in ir_1 * ir_2:

                    if irrep_normalization == "component":
                        alpha = ir_out.dim
                    elif irrep_normalization == "norm":
                        alpha = ir_1.dim * ir_2.dim
                    elif irrep_normalization == "none":
                        alpha = 1
                    else:
                        raise ValueError(f"irrep_normalization={irrep_normalization}")

                    if i_1 < i_2:
                        instructions += [(i_1, i_2, i_out, "uvw", True, alpha)]
                    elif i_1 == i_2:
                        i = i_1
                        mul = mul_1

                        if mul > 1:
                            instructions += [(i, i, i_out, "u<vw", True, alpha)]

                        if ir_out.l % 2 == 0:
                            if irrep_normalization == "component":
                                if ir_out.l == 0:
                                    alpha = ir_out.dim / (ir_1.dim + 2)
                                else:
                                    alpha = ir_out.dim / 2
                            if irrep_normalization == "norm":
                                if ir_out.l == 0:
                                    alpha = ir_out.dim * ir_1.dim
                                else:
                                    alpha = ir_1.dim * (ir_1.dim + 2) / 2

                            instructions += [(i, i, i_out, "uuw", True, alpha)]

    return FunctionalTensorProduct(irreps_in, irreps_in, irreps_out, instructions, irrep_normalization="none", **kwargs)


class TensorSquare(hk.Module):
    def __init__(self, irreps_out, *, irreps_in=None, init=None):
        super().__init__()

        self.irreps_in = Irreps(irreps_in) if irreps_in is not None else None
        self.irreps_out = Irreps(irreps_out)

        if init is None:
            init = hk.initializers.RandomNormal()
        self.init = init

    def __call__(self, input: IrrepsData) -> IrrepsData:
        if self.irreps_in is not None:
            input = IrrepsData.new(self.irreps_in, input)

        input = input.remove_nones().simplify()

        tp = FunctionalTensorSquare(input.irreps, self.irreps_out)
        ws = [
            hk.get_parameter(
                (
                    f"w[{ins.i_in1},{ins.i_in2},{ins.i_out}] "
                    f"{tp.irreps_in1[ins.i_in1]},{tp.irreps_in2[ins.i_in2]},{tp.irreps_out[ins.i_out]}"
                ),
                shape=ins.path_shape,
                init=self.init,
            )
            for ins in tp.instructions
        ]
        return tp.left_right(ws, input, input)



r"""Allows for clean lookup of Irreducible representations of :math:`O(3)`

Examples
--------
Create a scalar representation (:math:`l=0`) of even parity.

>>> from e3nn_jax import irrep
>>> irrep.l0e == Irrep("0e")
True

>>> from e3nn_jax.irrep import l1o, l2o
>>> l1o + l2o == Irrep("1o") + Irrep("2o")
True
"""

from .._irreps import Irrep


def __getattr__(name: str):
    r"""Creates an Irreps obeject by reflection

    Parameters
    ----------
    name : string
        the o3 object name prefixed by l. Example: l1o == Irrep("1o")

    Returns
    -------
    `e3nn_jax.Irrep`
        irreducible representation of :math:`O(3)`
    """

    prefix, *ir = name
    if prefix != "l" or not ir:
        raise AttributeError(f"'e3nn_jax.irrep' module has no attribute '{name}'")

    try:
        return Irrep("".join(ir))
    except (ValueError, AssertionError):
        raise AttributeError(f"'e3nn_jax.irrep' module has no attribute '{name}'")



import collections
import dataclasses
import itertools
import math
from functools import partial
from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.scipy

from e3nn_jax import axis_angle_to_angles, matrix_to_angles, perm, quaternion_to_angles, wigner_D


@dataclasses.dataclass(init=False, frozen=True)
class Irrep:
    r"""Irreducible representation of :math:`O(3)`

    This class does not contain any data, it is a structure that describe the representation.
    It is typically used as argument of other classes of the library to define the input and output
    representations of functions.

    Args:
        l: non-negative integer, the degree of the representation, :math:`l = 0, 1, \dots`
        p: {1, -1}, the parity of the representation

    Examples:
        Create a scalar representation (:math:`l=0`) of even parity.

        >>> Irrep(0, 1)
        0e

        Create a pseudotensor representation (:math:`l=2`) of odd parity.

        >>> Irrep(2, -1)
        2o

        Create a vector representation (:math:`l=1`) of the parity of the spherical harmonics (:math:`-1^l` gives odd parity).

        >>> Irrep("1y")
        1o

        >>> Irrep("2o").dim
        5

        >>> Irrep("2e") in Irrep("1o") * Irrep("1o")
        True

        >>> Irrep("1o") + Irrep("2o")
        1x1o+1x2o
    """
    l: int
    p: int

    def __init__(self, l, p=None):
        if p is None:
            if isinstance(l, Irrep):
                p = l.p
                l = l.l

            if isinstance(l, str):
                try:
                    name = l.strip()
                    l = int(name[:-1])
                    assert l >= 0
                    p = {
                        "e": 1,
                        "o": -1,
                        "y": (-1) ** l,
                    }[name[-1]]
                except Exception:
                    raise ValueError(f'unable to convert string "{name}" into an Irrep')
            elif isinstance(l, tuple):
                l, p = l

        assert isinstance(l, int) and l >= 0, l
        assert p in [-1, 1], p
        object.__setattr__(self, "l", l)
        object.__setattr__(self, "p", p)

    def __repr__(self):
        p = {+1: "e", -1: "o"}[self.p]
        return f"{self.l}{p}"

    @classmethod
    def iterator(cls, lmax=None):
        r"""Iterator through all the irreps of :math:`O(3)`

        Examples:
            >>> it = Irrep.iterator()
            >>> next(it), next(it), next(it), next(it)
            (0e, 0o, 1o, 1e)
        """
        for l in itertools.count():
            yield Irrep(l, (-1) ** l)
            yield Irrep(l, -((-1) ** l))

            if l == lmax:
                break

    def D_from_angles(self, alpha, beta, gamma, k=0):
        r"""Matrix :math:`p^k D^l(\alpha, \beta, \gamma)`

        (matrix) Representation of :math:`O(3)`. :math:`D` is the representation of :math:`SO(3)`, see `wigner_D`.

        Args:
            alpha (`jnp.ndarray`): of shape :math:`(...)`
                Rotation :math:`\alpha` around Y axis, applied third.
            beta (`jnp.ndarray`): of shape :math:`(...)`
                Rotation :math:`\beta` around X axis, applied second.
            gamma (`jnp.ndarray`): of shape :math:`(...)`
                Rotation :math:`\gamma` around Y axis, applied first.
            k (optional `jnp.ndarray`): of shape :math:`(...)`
                How many times the parity is applied.

        Returns:
            `jnp.ndarray`: of shape :math:`(..., 2l+1, 2l+1)`

        See Also:
            o3.wigner_D
            Irreps.D_from_angles
        """
        alpha, beta, gamma, k = jnp.broadcast_arrays(alpha, beta, gamma, k)
        k = jnp.asarray(k)
        return wigner_D(self.l, alpha, beta, gamma) * self.p ** k[..., None, None]

    def D_from_quaternion(self, q, k=0):
        r"""Matrix of the representation, see `Irrep.D_from_angles`

        Args:
            q (`jnp.ndarray`): shape :math:`(..., 4)`
            k (optional `jnp.ndarray`): shape :math:`(...)`

        Returns:
            `jnp.ndarray`: shape :math:`(..., 2l+1, 2l+1)`
        """
        return self.D_from_angles(*quaternion_to_angles(q), k)

    def D_from_matrix(self, R):
        r"""Matrix of the representation

        Args:
            R (`jnp.ndarray`): array of shape :math:`(..., 3, 3)`
            k (`jnp.ndarray`, optional): array of shape :math:`(...)`

        Returns:
            `jnp.ndarray`: array of shape :math:`(..., 2l+1, 2l+1)`

        Examples:
            >>> m = Irrep(1, -1).D_from_matrix(-jnp.eye(3))
            >>> m + 0.0
            DeviceArray([[-1.,  0.,  0.],
                         [ 0., -1.,  0.],
                         [ 0.,  0., -1.]], dtype=float32)

        See Also:
            `Irrep.D_from_angles`
        """
        d = jnp.sign(jnp.linalg.det(R))
        R = d[..., None, None] * R
        k = (1 - d) / 2
        return self.D_from_angles(*matrix_to_angles(R), k)

    @property
    def dim(self) -> int:
        """The dimension of the representation, :math:`2 l + 1`."""
        return 2 * self.l + 1

    def is_scalar(self) -> bool:
        """Equivalent to ``l == 0 and p == 1``"""
        return self.l == 0 and self.p == 1

    def __mul__(self, other):
        r"""Generate the irreps from the product of two irreps.

        Returns:
            generator of `Irrep`
        """
        other = Irrep(other)
        p = self.p * other.p
        lmin = abs(self.l - other.l)
        lmax = self.l + other.l
        for l in range(lmin, lmax + 1):
            yield Irrep(l, p)

    def __rmul__(self, other):
        r"""
        >>> 3 * Irrep('1e')
        3x1e
        """
        assert isinstance(other, int)
        return Irreps([(other, self)])

    def __add__(self, other):
        return Irreps(self) + Irreps(other)

    def __iter__(self):
        yield self.l
        yield self.p

    def __lt__(self, other):
        return (self.l, self.p) < (other.l, other.p)


jax.tree_util.register_pytree_node(Irrep, lambda ir: ((), ir), lambda ir, _: ir)


@dataclasses.dataclass(init=False, frozen=True)
class MulIrrep:
    mul: int
    ir: Irrep

    def __init__(self, mul, ir=None):
        if ir is None:
            mul, ir = mul

        object.__setattr__(self, "mul", mul)
        object.__setattr__(self, "ir", ir)

    @property
    def dim(self) -> int:
        return self.mul * self.ir.dim

    def __repr__(self):
        return f"{self.mul}x{self.ir}"

    def __iter__(self):
        yield self.mul
        yield self.ir

    def __lt__(self, other):
        return (self.mul, self.ir) < (other.mul, other.ir)


jax.tree_util.register_pytree_node(MulIrrep, lambda mulir: ((), mulir), lambda mulir, _: mulir)


class Irreps(tuple):
    r"""Direct sum of irreducible representations of :math:`O(3)`

    This class does not contain any data, it is a structure that describe the representation.
    It is typically used as argument of other classes of the library to define the input and output
    representations of functions.

    Attributes:
        dim (int): the total dimension of the representation
        num_irreps (int): number of irreps. the sum of the multiplicities
        ls (list of int): list of :math:`l` values
        lmax (int): maximum :math:`l` value

    Examples:
        >>> x = Irreps([(100, (0, 1)), (50, (1, 1))])
        >>> x
        100x0e+50x1e

        >>> x.dim
        250

        >>> Irreps("100x0e + 50x1e")
        100x0e+50x1e

        >>> Irreps("100x0e + 50x1e + 0x2e")
        100x0e+50x1e+0x2e

        >>> Irreps("100x0e + 50x1e + 0x2e").lmax
        1

        >>> Irrep("2e") in Irreps("0e + 2e")
        True

        Empty Irreps

        >>> Irreps(), Irreps("")
        (, )
    """

    def __new__(cls, irreps=None):
        if isinstance(irreps, Irreps):
            return super().__new__(cls, irreps)

        out = []
        if isinstance(irreps, Irrep):
            out.append(MulIrrep(1, Irrep(irreps)))
        elif isinstance(irreps, str):
            try:
                if irreps.strip() != "":
                    for mul_ir in irreps.split("+"):
                        if "x" in mul_ir:
                            mul, ir = mul_ir.split("x")
                            mul = int(mul)
                            ir = Irrep(ir)
                        else:
                            mul = 1
                            ir = Irrep(mul_ir)

                        assert isinstance(mul, int) and mul >= 0
                        out.append(MulIrrep(mul, ir))
            except Exception:
                raise ValueError(f'Unable to convert string "{irreps}" into an Irreps')
        elif irreps is None:
            pass
        else:
            for mul_ir in irreps:
                if isinstance(mul_ir, str):
                    mul = 1
                    ir = Irrep(mul_ir)
                elif isinstance(mul_ir, Irrep):
                    mul = 1
                    ir = mul_ir
                elif isinstance(mul_ir, MulIrrep):
                    mul, ir = mul_ir
                elif isinstance(mul_ir, int):
                    mul, ir = 1, Irrep(l=mul_ir, p=1)
                elif len(mul_ir) == 2:
                    mul, ir = mul_ir
                    ir = Irrep(ir)
                else:
                    mul = None
                    ir = None

                if not (isinstance(mul, int) and mul >= 0 and ir is not None):
                    raise ValueError(f'Unable to interpret "{mul_ir}" as an irrep.')

                out.append(MulIrrep(mul, ir))
        return super().__new__(cls, out)

    @staticmethod
    def spherical_harmonics(lmax, p=-1):
        r"""representation of the spherical harmonics

        Args:
            lmax (int): maximum :math:`l`
            p (optional {1, -1}): the parity of the representation

        Returns:
            `Irreps`: representation of :math:`(Y^0, Y^1, \dots, Y^{\mathrm{lmax}})`

        Examples:
            >>> Irreps.spherical_harmonics(3)
            1x0e+1x1o+1x2e+1x3o
            >>> Irreps.spherical_harmonics(4, p=1)
            1x0e+1x1e+1x2e+1x3e+1x4e
        """
        return Irreps([(1, (l, p**l)) for l in range(lmax + 1)])

    def slices(self):
        r"""List of slices corresponding to indices for each irrep.

        Examples:
            >>> Irreps('2x0e + 1e').slices()
            [slice(0, 2, None), slice(2, 5, None)]
        """
        s = []
        i = 0
        for mul_ir in self:
            s.append(slice(i, i + mul_ir.dim))
            i += mul_ir.dim
        return s

    def randn(self, key, size, *, normalization="component"):
        r"""Random tensor.

        Args:
            *size (list of int): size of the output tensor, needs to contains a ``-1``
            normalization : {'component', 'norm'}

        Returns:
            `jnp.ndarray`: array of shape ``size`` where ``-1`` is replaced by ``self.dim``

        Examples:
            >>> key = jax.random.PRNGKey(0)
            >>> Irreps("5x0e + 10x1o").randn(key, (5, -1, 5), normalization='norm').shape
            (5, 35, 5)

            >>> random_tensor = Irreps("2o").randn(key, (2, -1, 3), normalization='norm')
            >>> jnp.max(jnp.abs(jnp.linalg.norm(random_tensor, axis=1) - 1)) < 1e-5
            DeviceArray(True, dtype=bool)
        """
        di = size.index(-1)
        lsize = size[:di]
        rsize = size[di + 1 :]

        if normalization == "component":
            return jax.random.normal(key, lsize + (self.dim,) + rsize)
        elif normalization == "norm":
            x = jnp.zeros(lsize + (self.dim,) + rsize)
            for s, (mul, ir) in zip(self.slices(), self):
                key, k = jax.random.split(key)
                r = jax.random.normal(k, lsize + (mul, ir.dim) + rsize)
                r = r / jnp.linalg.norm(r, axis=di + 1, keepdims=True)
                i = di * (slice(None, None, None),) + (s,)
                x = x.at[i].set(r.reshape(*lsize, -1, *rsize))
            return x
        else:
            raise ValueError("Normalization needs to be 'norm' or 'component'")

    def __getitem__(self, i):
        x = super().__getitem__(i)
        if isinstance(i, slice):
            return Irreps(x)
        return x

    def __contains__(self, ir) -> bool:
        ir = Irrep(ir)
        return ir in (irrep for _, irrep in self)

    def count(self, ir) -> int:
        r"""Multiplicity of ``ir``.

        Args:
            ir (`Irrep`):

        Returns:
            `int`: total multiplicity of ``ir``
        """
        ir = Irrep(ir)
        return sum(mul for mul, irrep in self if ir == irrep)

    def index(self, _object):
        raise NotImplementedError

    def __add__(self, irreps):
        irreps = Irreps(irreps)
        return Irreps(super().__add__(irreps))

    def __mul__(self, other):
        r"""
        >>> (Irreps('2x1e') * 3).simplify()
        6x1e
        """
        if isinstance(other, Irreps):
            raise NotImplementedError("Use o3.TensorProduct for this, see the documentation")
        return Irreps(super().__mul__(other))

    def __rmul__(self, other):
        r"""
        >>> 2 * Irreps('0e + 1e')
        1x0e+1x1e+1x0e+1x1e
        """
        return Irreps(super().__rmul__(other))

    def unify(self):
        r"""Regroup same irrep together.

        Returns:
            `Irreps`: new `Irreps` object

        Examples:
            >>> Irreps('0e + 1e').unify()
            1x0e+1x1e

            >>> Irreps('0e + 1e + 1e').unify()
            1x0e+2x1e

            >>> Irreps('0e + 0x1e + 0e').unify()
            1x0e+0x1e+1x0e
        """
        out = []
        for mul, ir in self:
            if out and out[-1][1] == ir:
                out[-1] = (out[-1][0] + mul, ir)
            else:
                out.append((mul, ir))
        return Irreps(out)

    def remove_zero_multiplicities(self):
        """Remove any irreps with multiplicities of zero.

        Examples:
            >>> Irreps("4x0e + 0x1o + 2x3e").remove_zero_multiplicities()
            4x0e+2x3e
        """
        return Irreps([(mul, ir) for mul, ir in self if mul > 0])

    def simplify(self):
        """Simplify the representations.

        Examples:
            Note that simplify does not sort the representations.

            >>> Irreps("1e + 1e + 0e").simplify()
            2x1e+1x0e

            Equivalent representations which are separated from each other are not combined.

            >>> Irreps("1e + 1e + 0e + 1e").simplify()
            2x1e+1x0e+1x1e

            Except if they are separated by an irrep with multiplicity of zero.

            >>> Irreps("1e + 0x0e + 1e").simplify().simplify()
            2x1e
        """
        return self.remove_zero_multiplicities().unify()

    def sort(self):
        r"""Sort the representations.

        Returns:
            irreps (`Irreps`): sorted irreps
            p (tuple of int): permutation of the indices
            inv (tuple of int): inverse permutation of the indices

        Examples:
            >>> Irreps("1e + 0e + 1e").sort().irreps
            1x0e+1x1e+1x1e
            >>> Irreps("2o + 1e + 0e + 1e").sort().p
            (3, 1, 0, 2)
            >>> Irreps("2o + 1e + 0e + 1e").sort().inv
            (2, 1, 3, 0)
        """
        Ret = collections.namedtuple("sort", ["irreps", "p", "inv"])
        out = [(ir, i, mul) for i, (mul, ir) in enumerate(self)]
        out = sorted(out)
        inv = tuple(i for _, i, _ in out)
        p = perm.inverse(inv)
        irreps = Irreps([(mul, ir) for ir, _, mul in out])
        return Ret(irreps, p, inv)

    def filter(self, ir_list):
        r"""Filter the irreps.

        Args:
            ir_list (list of `Irrep`): list of irrep to keep

        Returns:
            `Irreps`: filtered irreps

        Examples:
            >>> Irreps("1e + 2e + 0e").filter(["0e", "1e"])
            1x1e+1x0e
        """
        if isinstance(ir_list, (str, Irrep)):
            ir_list = [ir_list]
        ir_list = {Irrep(ir) for ir in ir_list}
        return Irreps([(mul, ir) for mul, ir in self if ir in ir_list])

    @property
    def dim(self) -> int:
        return sum(mul * ir.dim for mul, ir in self)

    @property
    def num_irreps(self) -> int:
        return sum(mul for mul, _ in self)

    @property
    def ls(self) -> List[int]:
        return [l for mul, (l, p) in self for _ in range(mul)]

    @property
    def lmax(self) -> int:
        if len(self) == 0:
            raise ValueError("Cannot get lmax of empty Irreps")
        return max(self.ls)

    def __repr__(self):
        return "+".join(f"{mul_ir}" for mul_ir in self)

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def transform_by_angles(self, contiguous, alpha, beta, gamma, k=0):
        r"""Rotate the data by angles according to the irreps

        Args:
            contiguous (`jnp.ndarray`): data compatible with the irreps.
            alpha (float): third rotation angle around the second axis (in radians)
            beta (float): second rotation angle around the first axis (in radians)
            gamma (float): first rotation angle around the second axis (in radians)
            k (int): parity operation

        Returns:
            `jnp.ndarray`
        """
        assert contiguous.shape[-1] == self.dim
        return IrrepsData.from_contiguous(self, contiguous).transform_by_angles(alpha, beta, gamma, k).contiguous

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def transform_by_quaternion(self, contiguous, q, k=0):
        r"""Rotate data by a rotation given by a quaternion

        Args:
            contiguous (`jnp.ndarray`): data compatible with the irreps.
            q (`jnp.ndarray`): quaternion
            k (int): parity operation

        Returns:
            `jnp.ndarray`

        Examples:
            >>> irreps = Irreps("0e + 1o + 2e")
            >>> irreps.transform_by_quaternion(
            ...     jnp.array([1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
            ...     jnp.array([1.0, 0.0, 0.0, 0.0]),
            ...     1
            ... ) + 0.0
            DeviceArray([ 1.,  0., -1.,  0.,  0.,  0.,  1.,  0.,  0.], dtype=float32)
        """
        return self.transform_by_angles(contiguous, *quaternion_to_angles(q), k)

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def transform_by_axis_angle(self, contiguous, axis, angle, k=0):
        r"""Rotate data by an axis and an angle

        Args:
            contiguous (`jnp.ndarray`): data compatible with the irreps.
            axis (`jnp.ndarray`): axis
            angle (float): angle (in radians)
            k (int): parity operation

        Returns:
            `jnp.ndarray`
        """
        return self.transform_by_angles(contiguous, *axis_angle_to_angles(axis, angle), k)

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def transform_by_matrix(self, contiguous, R):
        r"""Rotate data by a rotation given by a matrix

        Args:
            contiguous (`jnp.ndarray`): data compatible with the irreps.
            R (`jnp.ndarray`): rotation matrix

        Returns:
            `jnp.ndarray`
        """
        d = jnp.sign(jnp.linalg.det(R))
        R = d[..., None, None] * R
        k = (1 - d) / 2
        return self.transform_by_angles(contiguous, *matrix_to_angles(R), k)

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def D_from_angles(self, alpha, beta, gamma, k=0):
        r"""Compute the D matrix from the angles

        Args:
            alpha (float): third rotation angle around the second axis (in radians)
            beta (float): second rotation angle around the first axis (in radians)
            gamma (float): first rotation angle around the second axis (in radians)
            k (int): parity operation

        Returns:
            `jnp.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
        """
        return jax.scipy.linalg.block_diag(*[ir.D_from_angles(alpha, beta, gamma, k) for mul, ir in self for _ in range(mul)])

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def D_from_quaternion(self, q, k=0):
        r"""Matrix of the representation

        Args:
            q (`jnp.ndarray`): array of shape :math:`(..., 4)`
            k (`jnp.ndarray`, optional): array of shape :math:`(...)`

        Returns:
            `jnp.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
        """
        return self.D_from_angles(*quaternion_to_angles(q), k)

    @partial(jax.jit, static_argnums=(0,), inline=True)
    def D_from_matrix(self, R):
        r"""Matrix of the representation

        Args:
            R (`jnp.ndarray`): array of shape :math:`(..., 3, 3)`

        Returns:
            `jnp.ndarray`: array of shape :math:`(..., \mathrm{dim}, \mathrm{dim})`
        """
        d = jnp.sign(jnp.linalg.det(R))
        R = d[..., None, None] * R
        k = (1 - d) / 2
        return self.D_from_angles(*matrix_to_angles(R), k)


jax.tree_util.register_pytree_node(Irreps, lambda irreps: ((), irreps), lambda irreps, _: irreps)


class IrrepsData:
    r"""Class storing data and its irreps"""

    irreps: Irreps
    contiguous: jnp.ndarray  # this field is mendatory because it contains the shape
    _list: List[Optional[jnp.ndarray]]  # this field is lazy, it is computed only when needed

    def __init__(self, irreps: Irreps, contiguous: jnp.ndarray, list: List[Optional[jnp.ndarray]] = None):
        self.irreps = Irreps(irreps)
        self.contiguous = contiguous
        self._list = list

    @property
    def list(self) -> List[Optional[jnp.ndarray]]:
        if self._list is None:
            shape = self.contiguous.shape[:-1]
            if len(self.irreps) == 1:
                mul, ir = self.irreps[0]
                list = [jnp.reshape(self.contiguous, shape + (mul, ir.dim))]
            else:
                list = [
                    jnp.reshape(self.contiguous[..., i], shape + (mul, ir.dim))
                    for i, (mul, ir) in zip(self.irreps.slices(), self.irreps)
                ]
            self._list = list
        return self._list

    @staticmethod
    def zeros(irreps: Irreps, shape) -> "IrrepsData":
        irreps = Irreps(irreps)
        return IrrepsData(irreps=irreps, contiguous=jnp.zeros(shape + (irreps.dim,)), list=[None] * len(irreps))

    @staticmethod
    def ones(irreps: Irreps, shape) -> "IrrepsData":
        irreps = Irreps(irreps)
        return IrrepsData(
            irreps=irreps,
            contiguous=jnp.ones(shape + (irreps.dim,)),
            list=[jnp.ones(shape + (mul, ir.dim)) for mul, ir in irreps],
        )

    @staticmethod
    def new(irreps: Irreps, any) -> "IrrepsData":
        r"""Create a new IrrepsData

        Args:
            irreps (`Irreps`): the irreps of the data
            any: the data

        Returns:
            `IrrepsData`
        """
        if isinstance(any, IrrepsData):
            return any.convert(irreps)
        if isinstance(any, list):
            shape = None
            for x in any:
                if x is not None:
                    shape = x.shape[:-2]
            if shape is None:
                raise ValueError("IrrepsData.new cannot infer shape from list of arrays")
            return IrrepsData.from_list(irreps, any, shape)
        return IrrepsData.from_contiguous(irreps, any)

    @staticmethod
    def from_list(irreps: Irreps, list, shape: Tuple[int]) -> "IrrepsData":
        r"""Create an IrrepsData from a list of arrays

        Args:
            irreps (Irreps): irreps
            list (list of optional `jnp.ndarray`): list of arrays

        Returns:
            IrrepsData
        """
        irreps = Irreps(irreps)
        assert len(irreps) == len(list), f"irreps has {len(irreps)} elements and list has {len(list)}"
        assert all(x is None or isinstance(x, jnp.ndarray) for x in list)
        assert all(x is None or x.shape == shape + (mul, ir.dim) for x, (mul, ir) in zip(list, irreps)), (
            [x.shape for x in list if x is not None],
            shape,
        )

        if irreps.dim > 0:
            contiguous = jnp.concatenate(
                [
                    jnp.zeros(shape + (mul_ir.dim,)) if x is None else x.reshape(shape + (mul_ir.dim,))
                    for mul_ir, x in zip(irreps, list)
                ],
                axis=-1,
            )
        else:
            contiguous = jnp.zeros(shape + (0,))
        return IrrepsData(irreps=irreps, contiguous=contiguous, list=list)

    @staticmethod
    def from_contiguous(irreps: Irreps, contiguous) -> "IrrepsData":
        r"""Create an IrrepsData from a contiguous array

        Args:
            irreps (Irreps): irreps
            contiguous (`jnp.ndarray`): contiguous array

        Returns:
            IrrepsData
        """
        assert contiguous.shape[-1] == Irreps(irreps).dim
        return IrrepsData(irreps=irreps, contiguous=contiguous, list=None)

    def __repr__(self):
        return f"IrrepsData({self.irreps}, {self.contiguous}, {self.list})"

    @property
    def shape(self):
        return self.contiguous.shape[:-1]

    @property
    def ndim(self):
        return len(self.shape)

    def reshape(self, shape) -> "IrrepsData":
        list = [None if x is None else x.reshape(shape + (mul, ir.dim)) for (mul, ir), x in zip(self.irreps, self.list)]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous.reshape(shape + (self.irreps.dim,)), list=list)

    def broadcast_to(self, shape) -> "IrrepsData":
        assert isinstance(shape, tuple)
        contiguous = jnp.broadcast_to(self.contiguous, shape + (self.irreps.dim,))
        list = [
            None if x is None else jnp.broadcast_to(x, shape + (mul, ir.dim)) for (mul, ir), x in zip(self.irreps, self.list)
        ]
        return IrrepsData(irreps=self.irreps, contiguous=contiguous, list=list)

    def replace_none_with_zeros(self) -> "IrrepsData":
        list = [jnp.zeros(self.shape + (mul, ir.dim)) if x is None else x for (mul, ir), x in zip(self.irreps, self.list)]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous, list=list)

    def remove_nones(self) -> "IrrepsData":
        if any(x is None for x in self.list):
            irreps = [mul_ir for mul_ir, x in zip(self.irreps, self.list) if x is not None]
            list = [x for x in self.list if x is not None]
            return IrrepsData.from_list(irreps, list, self.shape)
        return self

    def simplify(self) -> "IrrepsData":
        return self.convert(self.irreps.simplify())

    def split(self, indices: List[int]) -> List["IrrepsData"]:
        contiguous_parts = jnp.split(self.contiguous, [self.irreps[:i].dim for i in indices], axis=-1)
        assert len(contiguous_parts) == len(indices) + 1
        return [
            IrrepsData(irreps=self.irreps[i:j], contiguous=contiguous, list=self.list[i:j])
            for (i, j), contiguous in zip(zip([0] + indices, indices + [len(self.irreps)]), contiguous_parts)
        ]

    def repeat_irreps_by_last_axis(self) -> "IrrepsData":
        assert len(self.shape) >= 1
        irreps = (self.shape[-1] * self.irreps).simplify()
        contiguous = self.contiguous.reshape(self.shape[:-1] + (irreps.dim,))
        return IrrepsData.from_contiguous(irreps, contiguous)

    def repeat_mul_by_last_axis(self) -> "IrrepsData":
        assert len(self.shape) >= 1
        irreps = Irreps([(self.shape[-1] * mul, ir) for mul, ir in self.irreps])
        list = [None if x is None else x.reshape(self.shape[:-1] + (mul, ir.dim)) for (mul, ir), x in zip(irreps, self.list)]
        return IrrepsData.from_list(irreps, list, self.shape[:-1])

    def factor_irreps_to_last_axis(self) -> "IrrepsData":
        raise NotImplementedError

    def factor_mul_to_last_axis(self, factor=None) -> "IrrepsData":
        if factor is None:
            factor = math.gcd(*(mul for mul, _ in self.irreps))

        if not all(mul % factor == 0 for mul, _ in self.irreps):
            raise ValueError(f"factor {factor} does not divide all multiplicities")

        irreps = Irreps([(mul // factor, ir) for mul, ir in self.irreps])
        list = [
            None if x is None else x.reshape(self.shape + (factor, mul, ir.dim)) for (mul, ir), x in zip(irreps, self.list)
        ]
        return IrrepsData.from_list(irreps, list, self.shape + (factor,))

    def transform_by_angles(self, alpha: float, beta: float, gamma: float, k: int = 0) -> "IrrepsData":
        r"""Rotate the data by angles according to the irreps

        Args:
            alpha (float): third rotation angle around the second axis (in radians)
            beta (float): second rotation angle around the first axis (in radians)
            gamma (float): first rotation angle around the second axis (in radians)
            k (int): parity operation

        Returns:
            IrrepsData
        """
        # Optimization: we use only the list of arrays, not the contiguous data
        D = {ir: ir.D_from_angles(alpha, beta, gamma, k) for ir in {ir for _, ir in self.irreps}}
        new_list = [
            jnp.reshape(jnp.einsum("ij,...uj->...ui", D[ir], x), self.shape + (mul, ir.dim)) if x is not None else None
            for (mul, ir), x in zip(self.irreps, self.list)
        ]
        return IrrepsData.from_list(self.irreps, new_list, self.shape)

    def transform_by_quaternion(self, q: jnp.ndarray, k: int = 0) -> "IrrepsData":
        r"""Rotate data by a rotation given by a quaternion

        Args:
            q (`jnp.ndarray`): quaternion
            k (int): parity operation

        Returns:
            IrrepsData
        """
        return self.transform_by_angles(*quaternion_to_angles(q), k)

    def transform_by_axis_angle(self, axis: jnp.ndarray, angle: float, k: int = 0) -> "IrrepsData":
        r"""Rotate data by a rotation given by an axis and an angle

        Args:
            axis (`jnp.ndarray`): axis
            angle (float): angle (in radians)
            k (int): parity operation

        Returns:
            IrrepsData
        """
        return self.transform_by_angles(*axis_angle_to_angles(axis, angle), k)

    def transform_by_matrix(self, R: jnp.ndarray) -> "IrrepsData":
        r"""Rotate data by a rotation given by a matrix

        Args:
            R (`jnp.ndarray`): rotation matrix

        Returns:
            IrrepsData
        """
        d = jnp.sign(jnp.linalg.det(R))
        R = d[..., None, None] * R
        k = (1 - d) / 2
        return self.transform_by_angles(*matrix_to_angles(R), k)

    def convert(self, irreps: Irreps) -> "IrrepsData":
        r"""Convert the list property into an equivalent irreps

        Args:
            irreps (Irreps): new irreps

        Returns:
            `IrrepsData`: data with the new irreps

        Raises:
            ValueError: if the irreps are not compatible

        Example:
        >>> id = IrrepsData.new("10x0e + 10x0e", [None, jnp.ones((1, 10, 1))])
        >>> jax.tree_map(lambda x: x.shape, id.convert("20x0e")).list
        [(1, 20, 1)]
        """
        # Optimization: we use only the list of arrays, not the contiguous data
        irreps = Irreps(irreps)
        assert self.irreps.simplify() == irreps.simplify(), (self.irreps, irreps)
        # TODO test cases with mul == 0

        shape = self.shape

        new_list = []
        current_array = 0

        while len(new_list) < len(irreps) and irreps[len(new_list)].mul == 0:
            new_list.append(None)

        for mul_ir, y in zip(self.irreps, self.list):
            mul, _ = mul_ir

            while mul > 0:
                if isinstance(current_array, int):
                    current_mul = current_array
                else:
                    current_mul = current_array.shape[-2]

                needed_mul = irreps[len(new_list)].mul - current_mul

                if mul <= needed_mul:
                    x = y
                    m = mul
                    mul = 0
                elif mul > needed_mul:
                    if y is None:
                        x = None
                    else:
                        x, y = jnp.split(y, [needed_mul], axis=-2)
                    m = needed_mul
                    mul -= needed_mul

                if x is None:
                    if isinstance(current_array, int):
                        current_array += m
                    else:
                        current_array = jnp.concatenate([current_array, jnp.zeros(shape + (m, mul_ir.ir.dim))], axis=-2)
                else:
                    if isinstance(current_array, int):
                        if current_array == 0:
                            current_array = x
                        else:
                            current_array = jnp.concatenate([jnp.zeros(shape + (current_array, mul_ir.ir.dim)), x], axis=-2)
                    else:
                        current_array = jnp.concatenate([current_array, x], axis=-2)

                if isinstance(current_array, int):
                    if current_array == irreps[len(new_list)].mul:
                        new_list.append(None)
                        current_array = 0
                else:
                    if current_array.shape[-2] == irreps[len(new_list)].mul:
                        new_list.append(current_array)
                        current_array = 0

                while len(new_list) < len(irreps) and irreps[len(new_list)].mul == 0:
                    new_list.append(None)

        assert current_array == 0

        assert len(new_list) == len(irreps)
        assert all(x is None or isinstance(x, jnp.ndarray) for x in new_list)
        assert all(x is None or x.shape[-2:] == (mul, ir.dim) for x, (mul, ir) in zip(new_list, irreps))

        return IrrepsData(irreps=irreps, contiguous=self.contiguous, list=new_list)

    def __add__(self, other: "IrrepsData") -> "IrrepsData":
        assert self.irreps == other.irreps
        list = [x if y is None else (y if x is None else x + y) for x, y in zip(self.list, other.list)]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous + other.contiguous, list=list)

    def __sub__(self, other: "IrrepsData") -> "IrrepsData":
        assert self.irreps == other.irreps
        list = [x if y is None else (-y if x is None else x - y) for x, y in zip(self.list, other.list)]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous - other.contiguous, list=list)

    def __mul__(self, other) -> "IrrepsData":
        other = jnp.array(other)
        list = [None if x is None else x * other[..., None, None] for x in self.list]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous * other[..., None], list=list)

    def __rmul__(self, other) -> "IrrepsData":
        return self * other

    def __truediv__(self, other) -> "IrrepsData":
        other = jnp.array(other)
        list = [None if x is None else x / other[..., None, None] for x in self.list]
        return IrrepsData(irreps=self.irreps, contiguous=self.contiguous / other[..., None], list=list)

    @staticmethod
    def cat(args, axis="irreps"):
        r"""Concatenate IrrepsData

        Args:
            args (list of `IrrepsData`): list of data to concatenate
            axis (str or int): axis to concatenate on

        Returns:
            `IrrepsData`: concatenated data
        """
        assert len(args) >= 1
        if axis == "irreps":
            irreps = Irreps(sum([x.irreps for x in args], Irreps("")))
            return IrrepsData(
                irreps=irreps,
                contiguous=jnp.concatenate([x.contiguous for x in args], axis=-1),
                list=sum([x.list for x in args], []),
            )
        elif axis == "mul":
            raise NotImplementedError

        assert isinstance(axis, int)
        assert {x.irreps for x in args} == {args[0].irreps}
        while axis < 0:
            axis += len(args[0].shape)
        args = [x.replace_none_with_zeros() for x in args]  # TODO this could be optimized
        return IrrepsData(
            irreps=args[0].irreps,
            contiguous=jnp.concatenate([x.contiguous for x in args], axis=axis),
            list=[jnp.concatenate(xs, axis=axis) for xs in zip(*[x.list for x in args])],
        )

    @staticmethod
    def randn(irreps, key, size=(), *, normalization="component"):
        irreps = Irreps(irreps)
        x = irreps.randn(key, size + (-1,), normalization=normalization)
        return IrrepsData.from_contiguous(irreps, x)


jax.tree_util.register_pytree_node(
    IrrepsData,
    lambda x: ((x.contiguous, x.list), x.irreps),
    lambda x, data: IrrepsData(irreps=x, contiguous=data[0], list=data[1]),
)



from fractions import Fraction
import numpy as np

import sympy


def Q_to_sympy(x: float) -> sympy.Expr:
    x = Fraction(x).limit_denominator()
    return sympy.sympify(x)


def sqrtQ_to_sympy(x: float) -> sympy.Expr:
    sign = 1 if x >= 0 else -1
    return sign * sympy.sqrt(Q_to_sympy(x**2))


def sqrtQarray_to_sympy(x: np.ndarray) -> sympy.Array:
    if x.ndim == 0:
        return sqrtQ_to_sympy(x)
    else:
        return sympy.Array([sqrtQarray_to_sympy(row) for row in x])



import jax
import jax.numpy as jnp


def index_add(i, x, out_dim):
    r"""perform the operation

    ```
    out = zeros(out_dim, ...)
    out[i] += x
    ```

    Args:
        i (`jnp.ndarray`): array of indices
        x (`jnp.ndarray`): array of data
        out_dim (int): size of the output

    Returns:
        `jnp.ndarray`: ``out``

    Example:
       >>> i = jnp.array([0, 2, 2, 0])
       >>> x = jnp.array([1.0, 2.0, 3.0, -10.0])
       >>> index_add(i, x, out_dim=4)
       DeviceArray([-9.,  0.,  5.,  0.], dtype=float32)
    """
    # out_dim = jnp.max(i) + 1
    return jnp.zeros((out_dim,) + x.shape[1:]).at[i].add(x)


def radius_graph(pos, r_max, *, batch=None, size=None, loop=False):
    r"""naive and inefficient version of ``torch_cluster.radius_graph``

    Args:
        pos (`jnp.ndarray`): array of shape ``(n, 3)``
        r_max (float):
        batch (`jnp.ndarray`): indices
        size (int): size of the output
        loop (bool): whether to include self-loops

    Returns:
        `jnp.ndarray`: src
        `jnp.ndarray`: dst

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> pos = jax.random.normal(key, (20, 3))
        >>> batch = jnp.arange(20) < 10
        >>> radius_graph(pos, 0.8, batch=batch)
        (DeviceArray([ 3,  7, 10, 11, 12, 18], dtype=int32), DeviceArray([ 7,  3, 11, 10, 18, 12], dtype=int32))
    """
    r = jax.vmap(jax.vmap(lambda x, y: jnp.linalg.norm(x - y), (None, 0), 0), (0, None), 0)(pos, pos)
    if loop:
        mask = r < r_max
    else:
        mask = (r < r_max) & (r > 0)

    src, dst = jnp.where(mask, size=size, fill_value=-1)

    if batch is None:
        return src, dst
    return src[batch[src] == batch[dst]], dst[batch[src] == batch[dst]]
